#############################################################################
#
# Copyright (c) 2006-2007 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Grokkers for the various components."""

import os

import zope.component.interface
from zope import interface, component
from zope.security.checker import NamesChecker
from zope.publisher.interfaces.browser import (IDefaultBrowserLayer,
                                               IBrowserRequest,
                                               IBrowserPublisher,
                                               IBrowserSkinType)
from zope.publisher.interfaces.xmlrpc import IXMLRPCRequest
from zope.security.interfaces import IPermission
from zope.app.securitypolicy.interfaces import IRole
from zope.app.securitypolicy.rolepermission import rolePermissionManager

from zope.annotation.interfaces import IAnnotations

from zope.app.publisher.xmlrpc import MethodPublisher
from zope.app.container.interfaces import IContainer
from zope.app.container.interfaces import INameChooser
from zope.app.container.contained import contained

from zope.app.intid import IntIds
from zope.app.intid.interfaces import IIntIds
from zope.app.catalog.catalog import Catalog
from zope.app.catalog.interfaces import ICatalog

from zope.exceptions.interfaces import DuplicationError

import martian
from martian.error import GrokError
from martian import util

import grok
from grok import components, formlib, templatereg
from grok.util import check_adapts, get_default_permission, make_checker
from grok.util import determine_class_directive, public_methods_from_class
from grok.rest import RestPublisher
from grok.interfaces import IRESTSkinType

class ContextGrokker(martian.GlobalGrokker):

    priority = 1001

    def grok(self, name, module, module_info, config, **kw):
        possible_contexts = martian.scan_for_classes(module, [grok.Model,
                                                              grok.Container])
        context = util.determine_module_context(module_info, possible_contexts)
        module.__grok_context__ = context
        return True


class AdapterGrokker(martian.ClassGrokker):
    component_class = grok.Adapter

    def grok(self, name, factory, module_info, config, **kw):
        context = module_info.getAnnotation('grok.context', None)
        adapter_context = util.determine_class_context(factory, context)
        provides = util.class_annotation(factory, 'grok.provides', None)
        if provides is None:
            util.check_implements_one(factory)
        name = util.class_annotation(factory, 'grok.name', '')

        config.action( 
            discriminator=('adapter', adapter_context, provides, name),
            callable=component.provideAdapter,
            args=(factory, (adapter_context,), provides, name),
            )
        return True

class MultiAdapterGrokker(martian.ClassGrokker):
    component_class = grok.MultiAdapter

    def grok(self, name, factory, module_info, config, **kw):
        provides = util.class_annotation(factory, 'grok.provides', None)
        if provides is None:
            util.check_implements_one(factory)
        check_adapts(factory)
        name = util.class_annotation(factory, 'grok.name', '')
        for_ = component.adaptedBy(factory)

        config.action(
            discriminator=('adapter', for_, provides, name),
            callable=component.provideAdapter,
            args=(factory, None, provides, name),
            )
        return True


class GlobalUtilityGrokker(martian.ClassGrokker):
    component_class = grok.GlobalUtility

    # This needs to happen before the FilesystemPageTemplateGrokker grokker 
    # happens, since it relies on the ITemplateFileFactories being grokked.
    priority = 1100

    def grok(self, name, factory, module_info, config, **kw):
        provides = util.class_annotation(factory, 'grok.provides', None)
        if provides is None:
            util.check_implements_one(factory)
        name = util.class_annotation(factory, 'grok.name', '')
        direct = util.class_annotation(factory, 'grok.direct', False)
        if not direct:
            factory = factory()

        config.action(
            discriminator=('utility', provides, name),
            callable=component.provideUtility,
            args=(factory, provides, name),
            )
        return True


class XMLRPCGrokker(martian.ClassGrokker):
    component_class = grok.XMLRPC

    def grok(self, name, factory, module_info, config, **kw):
        context = module_info.getAnnotation('grok.context', None)
        view_context = util.determine_class_context(factory, context)

        methods = public_methods_from_class(factory)

        default_permission = get_default_permission(factory)

        for method in methods:
            name = method.__name__
            if name.startswith('__'):
                continue

            # Make sure that the class inherits MethodPublisher, so that the
            # views have a location
            method_view = type(
                factory.__name__, (factory, MethodPublisher),
                {'__call__': method}
                )

            adapts = (view_context, IXMLRPCRequest)
            config.action(
                discriminator=('adapter', adapts, interface.Interface, name),
                callable=component.provideAdapter,
                args=(method_view, adapts, interface.Interface, name),
                )

            # Protect method_view with either the permission that was
            # set on the method, the default permission from the class
            # level or zope.Public.
            permission = getattr(method, '__grok_require__',
                                 default_permission)

            config.action(
                discriminator=('protectName', method_view, '__call__'),
                callable=make_checker,
                args=(factory, method_view, permission),
                )
        return True

class RESTGrokker(martian.ClassGrokker):
    component_class = grok.REST
    
    def grok(self, name, factory, module_info, config, **kw):
        context = module_info.getAnnotation('grok.context', None)
        view_context = util.determine_class_context(factory, context)

        methods = public_methods_from_class(factory)

        default_permission = get_default_permission(factory)

        # grab layer from class or module
        view_layer = determine_class_directive('grok.layer', factory,
                                               module_info,
                                               default=grok.IRESTLayer)
        
        for method in methods:
            name = method.__name__
            if name.startswith('__'):
                continue

            # Make sure that the class inherits RestPublisher, so that the
            # views have a location
            method_view = type(
                factory.__name__, (factory, RestPublisher),
                {'__call__': method }
                )

            adapts = (view_context, view_layer)
            config.action(
                discriminator=('adapter', adapts, interface.Interface, name),
                callable=component.provideAdapter,
                args=(method_view, adapts, interface.Interface, name),
                )

            # Protect method_view with either the permission that was
            # set on the method, the default permission from the class
            # level or zope.Public.
            permission = getattr(method, '__grok_require__',
                                 default_permission)
            config.action(
                discriminator=('protectName', method_view, '__call__'),
                callable=make_checker,
                args=(factory, method_view, permission),
                )
        return True
    
    
class ViewGrokker(martian.ClassGrokker):
    component_class = grok.View

    def grok(self, name, factory, module_info, config, **kw):
        context = module_info.getAnnotation('grok.context', None)
        view_context = util.determine_class_context(factory, context)

        factory.module_info = module_info
        factory_name = factory.__name__.lower()

        if util.check_subclass(factory, components.GrokForm):
            # setup form_fields from context class if we've encountered a form
            if getattr(factory, 'form_fields', None) is None:
                factory.form_fields = formlib.get_auto_fields(view_context)

            if not getattr(factory.render, 'base_method', False):
                raise GrokError(
                    "It is not allowed to specify a custom 'render' "
                    "method for form %r. Forms either use the default "
                    "template or a custom-supplied one." % factory,
                    factory)

        # find templates
        templates = module_info.getAnnotation('grok.templates', None)
        if templates is not None:
            config.action(
                discriminator=None,
                callable=templates.checkTemplates,
                args=(module_info, factory, factory_name)
            )

        # safety belt: make sure that the programmer didn't use
        # @grok.require on any of the view's methods.
        methods = util.methods_from_class(factory)
        for method in methods:
            if getattr(method, '__grok_require__', None) is not None:
                raise GrokError('The @grok.require decorator is used for '
                                'method %r in view %r. It may only be used '
                                'for XML-RPC methods.'
                                % (method.__name__, factory), factory)

        # grab layer from class or module
        view_layer = determine_class_directive('grok.layer',
                                               factory, module_info,
                                               default=IDefaultBrowserLayer)

        view_name = util.class_annotation(factory, 'grok.name',
                                          factory_name)
        # __view_name__ is needed to support IAbsoluteURL on views
        factory.__view_name__ = view_name
        adapts = (view_context, view_layer)

        config.action(
            discriminator=('adapter', adapts, interface.Interface, view_name),
            callable=component.provideAdapter,
            args=(factory, adapts, interface.Interface, view_name),
            )

        permission = get_default_permission(factory)
        config.action(
            discriminator=('protectName', factory, '__call__'),
            callable=make_checker,
            args=(factory, factory, permission),
            )

        return True

def view__getitem__(self, key):
    return self.template.macros[key]


class JSONGrokker(martian.ClassGrokker):
    component_class = grok.JSON

    def grok(self, name, factory, module_info, config, **kw):
        context = module_info.getAnnotation('grok.context', None)
        view_context = util.determine_class_context(factory, context)

        methods = public_methods_from_class(factory)

        default_permission = get_default_permission(factory)

        for method in methods:
            # The grok.JSON component inherits methods from its baseclass
            # (being zope.publisher.browser.BrowserPage) with names that
            # do not start with an underscore, but should still not
            # be registered as views. Ignore these methods:
            if method.__name__ in ['browserDefault', 'publishTraverse']:
                continue

            # Create a new class with a __view_name__ attribute so the
            # JSON class knows what method to call.
            method_view = type(
                factory.__name__, (factory,),
                {'__view_name__': method.__name__}
                )
            adapts = (view_context, IDefaultBrowserLayer)
            name = method.__name__

            config.action(
                discriminator=('adapter', adapts, interface.Interface, name),
                callable=component.provideAdapter,
                args=(method_view, adapts, interface.Interface, name),
                )

            # Protect method_view with either the permission that was
            # set on the method, the default permission from the class
            # level or zope.Public.

            permission = getattr(method, '__grok_require__',
                                 default_permission)

            config.action(
                discriminator=('protectName', method_view, '__call__'),
                callable=make_checker,
                args=(factory, method_view, permission),
                )
        return True


class TraverserGrokker(martian.ClassGrokker):
    component_class = grok.Traverser

    def grok(self, name, factory, module_info, config, **kw):
        context = module_info.getAnnotation('grok.context', None)
        factory_context = util.determine_class_context(factory, context)
        adapts = (factory_context, IBrowserRequest)

        config.action(
            discriminator=('adapter', adapts, IBrowserPublisher, ''),
            callable=component.provideAdapter,
            args=(factory, adapts, IBrowserPublisher),
            )            
        return True


class TemplateGrokker(martian.GlobalGrokker):
    # this needs to happen before any other grokkers execute that use
    # the template registry
    priority = 1001

    def grok(self, name, module, module_info, config, **kw):
        module.__grok_templates__ = templatereg.TemplateRegistry()
        return True


class ModulePageTemplateGrokker(martian.InstanceGrokker):
    # this needs to happen before any other grokkers execute that actually
    # use the templates
    priority = 1000

    component_class = grok.components.BaseTemplate

    def grok(self, name, instance, module_info, config, **kw):
        templates = module_info.getAnnotation('grok.templates', None)
        if templates is None:
            return False
        config.action(
            discriminator=None,
            callable=templates.register,
            args=(name, instance)
        )
        config.action(
            discriminator=None,
            callable=instance._annotateGrokInfo,
            args=(name, module_info.dotted_name)
        )
        return True
        

class FilesystemPageTemplateGrokker(martian.GlobalGrokker):
    # do this early on, but after ModulePageTemplateGrokker, as
    # findFilesystem depends on module-level templates to be
    # already grokked for error reporting
    priority = 999

    def grok(self, name, module, module_info, config, **kw):
        templates = module_info.getAnnotation('grok.templates', None)
        if templates is None:
            return False
        config.action(
            discriminator=None,
            callable=templates.findFilesystem,
            args=(module_info,)
        )
        return True


class UnassociatedTemplatesGrokker(martian.GlobalGrokker):
    priority = -1001

    def grok(self, name, module, module_info, config, **kw):
        templates = module_info.getAnnotation('grok.templates', None)
        if templates is None:
            return False
        
        config.action(
            discriminator=None,
            callable=templates.checkUnassociated,
            args=(module_info,)
        )
        return True


class SubscriberGrokker(martian.GlobalGrokker):

    def grok(self, name, module, module_info, config, **kw):
        subscribers = module_info.getAnnotation('grok.subscribers', [])

        for factory, subscribed in subscribers:
            config.action( 
                discriminator=None,
                callable=component.provideHandler,
                args=(factory, subscribed),
                )

            for iface in subscribed:
                config.action(
                    discriminator=None,
                    callable=zope.component.interface.provideInterface,
                    args=('', iface)
                    )
        return True


class AdapterDecoratorGrokker(martian.GlobalGrokker):

    def grok(self, name, module, module_info, config, **kw):
        context = module_info.getAnnotation('grok.context', None)
        implementers = module_info.getAnnotation('implementers', [])
        for function in implementers:
            interfaces = getattr(function, '__component_adapts__', None)
            if interfaces is None:
                # There's no explicit interfaces defined, so we assume the
                # module context to be the thing adapted.
                util.check_context(module_info.getModule(), context)
                interfaces = (context, )
            
            config.action( 
                discriminator=('adapter', interfaces, function.__implemented__),
                callable=component.provideAdapter,
                args=(function, interfaces, function.__implemented__),
                )
        return True


allowed_resource_names = (
    'GET', 'HEAD', 'publishTraverse', 'browserDefault', 'request', '__call__')

allowed_resourcedir_names = allowed_resource_names + ('__getitem__', 'get')

class StaticResourcesGrokker(martian.GlobalGrokker):

    def grok(self, name, module, module_info, config, **kw):
        # we're only interested in static resources if this module
        # happens to be a package
        if not module_info.isPackage():
            return False

        resource_path = module_info.getResourcePath('static')
        if os.path.isdir(resource_path):
            static_module = module_info.getSubModuleInfo('static')
            if static_module is not None:
                if static_module.isPackage():
                    raise GrokError(
                        "The 'static' resource directory must not "
                        "be a python package.",
                        module_info.getModule())
                else:
                    raise GrokError(
                        "A package can not contain both a 'static' "
                        "resource directory and a module named "
                        "'static.py'", module_info.getModule())

        checker = NamesChecker(allowed_resourcedir_names)
        resource_factory = components.DirectoryResourceFactory(
            resource_path, checker, module_info.dotted_name)

        adapts = (IDefaultBrowserLayer,)
        provides = interface.Interface
        name = module_info.dotted_name
        config.action(
            discriminator=('adapter', adapts, provides, name),
            callable=component.provideAdapter,
            args=(resource_factory, adapts, provides, name),
            )
        return True


class GlobalUtilityDirectiveGrokker(martian.GlobalGrokker):

    def grok(self, name, module, module_info, config, **kw):
        infos = module_info.getAnnotation('grok.global_utility', [])

        for info in infos:
            if info.provides is None:
                util.check_implements_one(info.factory)
            if info.direct:
                obj = info.factory
            else:
                obj = info.factory()
            component.provideUtility(obj,
                                     provides=info.provides,
                                     name=info.name)
        return True


class SiteGrokker(martian.ClassGrokker):
    component_class = grok.Site
    priority = 500

    def grok(self, name, factory, module_info, config, **kw):
        infos = util.class_annotation_list(factory, 'grok.local_utility', None)
        if infos is None:
            return False

        for info in infos:
            if info.public and not IContainer.implementedBy(factory):
                raise GrokError(
                    "Cannot set public to True with grok.local_utility as "
                    "the site (%r) is not a container." %
                    factory, factory)
            if info.provides is None:
                if util.check_subclass(info.factory, grok.LocalUtility):
                    baseInterfaces = interface.implementedBy(grok.LocalUtility)
                    utilityInterfaces = interface.implementedBy(info.factory)
                    provides = list(utilityInterfaces - baseInterfaces)

                    if len(provides) == 0 and len(list(utilityInterfaces)) > 0:
                        raise GrokError(
                            "Cannot determine which interface to use "
                            "for utility registration of %r in site %r. "
                            "It implements an interface that is a specialization "
                            "of an interface implemented by grok.LocalUtility. "
                            "Specify the interface by either using grok.provides "
                            "on the utility or passing 'provides' to "
                            "grok.local_utility." % (info.factory, factory),
                            info.factory)
                else:
                    provides = list(interface.implementedBy(info.factory))

                util.check_implements_one_from_list(provides, info.factory)
                info.provides = provides[0]

        # raise an error in case of any duplicate registrations
        # on the class level (subclassing overrides, see below)
        used = set()
        class_infos = util.class_annotation(factory, 'grok.local_utility',
                                            [])
        for info in class_infos:
            key = (info.provides, info.name)
            if key in used:
                raise GrokError(
                    "Conflicting local utility registration %r in "
                    "site %r. Local utilities are registered multiple "
                    "times for interface %r and name %r." %
                    (info.factory, factory, info.provides, info.name),
                    factory)
            used.add(key)

        # Make sure that local utilities from subclasses override
        # utilities from base classes if the registration (provided
        # interface, name) is identical.
        overridden_infos = []
        used = set()
        for info in reversed(infos):
            key = (info.provides, info.name)
            if key in used:
                continue
            used.add(key)
            overridden_infos.append(info)
        overridden_infos.reverse()

        # store infos on site class
        factory.__grok_utilities_to_install__ = overridden_infos
        adapts = (factory, grok.IObjectAddedEvent)

        config.action( 
            discriminator=None,
            callable=component.provideHandler,
            args=(localUtilityRegistrationSubscriber, adapts),
            )
        return True


def localUtilityRegistrationSubscriber(site, event):
    """A subscriber that fires to set up local utilities.
    """
    installed = getattr(site, '__grok_utilities_installed__', False)
    if installed:
        return

    for info in util.class_annotation(site.__class__,
                                      'grok.utilities_to_install', []):
        setupUtility(site, info.factory(), info.provides, name=info.name,
                     name_in_container=info.name_in_container,
                     public=info.public, setup=info.setup)

    # we are done. If this subscriber gets fired again, we therefore
    # do not register utilities anymore
    site.__grok_utilities_installed__ = True


def setupUtility(site, utility, provides, name=u'',
                 name_in_container=None, public=False, setup=None):
    """Set up a utility in a site.

    site - the site to set up the utility in
    utility - the utility to set up
    provides - the interface the utility should be registered with
    name - the name the utility should be registered under, default
      the empty string (no name)
    name_in_container - if given it will be used to add the utility
      object to its container. Otherwise a name will be made up
    public - if False, the utility will be stored in the site manager. If
      True, the utility will be storedin the site (it is assumed the
      site is a container)
    setup - if not None, it will be called with the utility as its first
       argument. This function can then be used to further set up the
       utility.
    """
    site_manager = site.getSiteManager()

    if not public:
        container = site_manager
    else:
        container = site

    if name_in_container is None:
        name_in_container = INameChooser(container).chooseName(
            utility.__class__.__name__, utility)
    container[name_in_container] = utility

    if setup is not None:
        setup(utility)

    site_manager.registerUtility(utility, provided=provides,
                                 name=name)

class PermissionGrokker(martian.ClassGrokker):
    component_class = grok.Permission
    priority = 1500

    def grok(self, name, factory, module_info, config, **kw):
        id = util.class_annotation(factory, 'grok.name', None)
        if id is None:
            raise GrokError(
                "A permission needs to have a dotted name for its id. Use "
                "grok.name to specify one.", factory)
        # We can safely convert to unicode, since the directives make sure
        # it is either unicode already or ASCII.
        id = unicode(id)
        permission = factory(
            id,
            unicode(util.class_annotation(factory, 'grok.title', id)),
            unicode(util.class_annotation(factory, 'grok.description', '')))

        config.action( 
            discriminator=('utility', IPermission, name),
            callable=component.provideUtility,
            args=(permission, IPermission, id),
            order=self.priority
            )
        return True

class RoleGrokker(martian.ClassGrokker):
    component_class = grok.Role
    priority = PermissionGrokker.priority - 1

    def grok(self, name, factory, module_info, config, **kw):
        id = util.class_annotation(factory, 'grok.name', None)
        if id is None:
            raise GrokError(
                "A role needs to have a dotted name for its id. Use "
                "grok.name to specify one.", factory)
        # We can safely convert to unicode, since the directives makes sure
        # it is either unicode already or ASCII.
        id = unicode(id)
        role = factory(
            id,
            unicode(util.class_annotation(factory, 'grok.title', id)),
            unicode(util.class_annotation(factory, 'grok.description', '')))

        config.action( 
            discriminator=('utility', IRole, name),
            callable=component.provideUtility,
            args=(role, IRole, id),
            )

        permissions = util.class_annotation(factory, 'grok.permissions', ())
        for permission in permissions:
            config.action(
                discriminator=('grantPermissionToRole', permission, id),
                callable=rolePermissionManager.grantPermissionToRole,
                args=(permission, id),
                )
        return True

class AnnotationGrokker(martian.ClassGrokker):
    component_class = grok.Annotation

    def grok(self, name, factory, module_info, config, **kw):
        context = module_info.getAnnotation('grok.context', None)
        adapter_context = util.determine_class_context(factory, context)
        provides = util.class_annotation(factory, 'grok.provides', None)
        if provides is None:
            base_interfaces = interface.implementedBy(grok.Annotation)
            factory_interfaces = interface.implementedBy(factory)
            real_interfaces = list(factory_interfaces - base_interfaces)
            util.check_implements_one_from_list(real_interfaces, factory)
            provides = real_interfaces[0]

        key = util.class_annotation(factory, 'grok.name', None)
        if key is None:
            key = factory.__module__ + '.' + factory.__name__

        @component.adapter(adapter_context)
        @interface.implementer(provides)
        def getAnnotation(context):
            annotations = IAnnotations(context)
            try:
                result = annotations[key]
            except KeyError:
                result = factory()
                annotations[key] = result

            # Containment has to be set up late to allow containment
            # proxies to be applied, if needed. This does not trigger
            # an event and is idempotent if containment is set up
            # already.
            contained_result = contained(result, context, key)
            return contained_result

        config.action(
            discriminator=('adapter', adapter_context, provides, ''),
            callable=component.provideAdapter,
            args=(getAnnotation,),
            )
        return True


class ApplicationGrokker(martian.ClassGrokker):
    component_class = grok.Application
    priority = 500

    def grok(self, name, factory, module_info, config, **kw):
        # XXX fail loudly if the same application name is used twice.
        provides = grok.interfaces.IApplication
        name = '%s.%s' % (module_info.dotted_name, name)
        config.action(
            discriminator=('utility', provides, name),
            callable=component.provideUtility,
            args=(factory, provides, name),
            )
        return True


class IndexesGrokker(martian.InstanceGrokker):
    component_class = components.IndexesClass

    def grok(self, name, factory, module_info, config, **kw):
        site = util.class_annotation(factory, 'grok.site', None)
        if site is None:
            raise GrokError("No site specified for grok.Indexes "
                            "subclass in module %r. "
                            "Use grok.site() to specify." % module_info.getModule(),
                            factory)
        indexes = util.class_annotation(factory, 'grok.indexes', None)
        if indexes is None:
            return False
        context = module_info.getAnnotation('grok.context', None)
        context = util.determine_class_context(factory, context)
        catalog_name = util.class_annotation(factory, 'grok.name', u'')

        subscriber = IndexesSetupSubscriber(catalog_name, indexes,
                                            context, module_info)
        subscribed = (site, grok.IObjectAddedEvent)
        config.action( 
            discriminator=None,
            callable=component.provideHandler,
            args=(subscriber, subscribed),
            )
        return True


class IndexesSetupSubscriber(object):

    def __init__(self, catalog_name, indexes, context, module_info):
        self.catalog_name = catalog_name
        self.indexes = indexes
        self.context = context
        self.module_info = module_info

    def __call__(self, site, event):
        # make sure we have an intids
        self._createIntIds(site)
        # get the catalog
        catalog = self._createCatalog(site)
        # now install indexes
        for name, index in self.indexes.items():
            try:
                index.setup(catalog, name, self.context, self.module_info)
            except DuplicationError:
                raise GrokError(
                    "grok.Indexes in module %r causes "
                    "creation of catalog index %r in catalog %r, "
                    "but an index with that name is already present." %
                    (self.module_info.getModule(), name, self.catalog_name),
                    None)

    def _createCatalog(self, site):
        """Create the catalog if needed and return it.

        If the catalog already exists, return that.

        """
        catalog = zope.component.queryUtility(
            ICatalog, name=self.catalog_name, context=site, default=None)
        if catalog is not None:
            return catalog
        catalog = Catalog()
        setupUtility(site, catalog, ICatalog, name=self.catalog_name)
        return catalog

    def _createIntIds(self, site):
        """Create intids if needed, and return it.
        """
        intids = zope.component.queryUtility(
            IIntIds, context=site, default=None)
        if intids is not None:
            return intids
        intids = IntIds()
        setupUtility(site, intids, IIntIds)
        return intids


class SkinGrokker(martian.ClassGrokker):
    component_class = grok.Skin

    def grok(self, name, factory, module_info, config, **kw):
        layer = determine_class_directive('grok.layer', factory, module_info,
                                          default=IBrowserRequest)
        name = grok.util.class_annotation(factory, 'grok.name',
                                          factory.__name__.lower())
        config.action(
            discriminator=None,
            callable=zope.component.interface.provideInterface,
            args=(name, layer, IBrowserSkinType)
            )
        return True

class RESTProtocolGrokker(martian.ClassGrokker):
    component_class = grok.RESTProtocol

    def grok(self, name, factory, module_info, config, **kw):
        layer = determine_class_directive('grok.layer', factory, module_info,
                                          default=IBrowserRequest)
        name = grok.util.class_annotation(factory, 'grok.name',
                                          factory.__name__.lower())
        config.action(
            discriminator=None,
            callable=zope.component.interface.provideInterface,
            args=(name, layer, IRESTSkinType)
            )
        return True



"""
If a form does not have a template, a simple default template is
associated with them. Otherwise, the supplied template is used.

  >>> from zope.publisher.browser import TestRequest
  >>> request = TestRequest()
  >>> from zope import component
  
Default edit template:

  >>> view = component.getMultiAdapter((Mammoth(), request), name='edit')
  >>> print view()
  <html>...
  
Custom edit template:

  >>> view = component.getMultiAdapter((Mammoth(), request), name='edit2')
  >>> print view()
  <p>Test edit</p>

Custom edit template with an explicit template

  >>> view = component.getMultiAdapter((Mammoth(), request), name='edit3')
  >>> print view()
  <p>Test edit</p>

Default display template:

  >>> view = component.getMultiAdapter((Mammoth(), request), name='display')
  >>> print view()
  <html>...
  
Custom display template:

  >>> view = component.getMultiAdapter((Mammoth(), request), name='display2')
  >>> print view()
  <p>Test display</p>

Custom display template with an explicit template:

  >>> view = component.getMultiAdapter((Mammoth(), request), name='display3')
  >>> print view()
  <p>Test display</p>

"""
import grok
from zope import schema

class Mammoth(grok.Model):
    class fields:
        name = schema.TextLine(title=u"Name")
        size = schema.TextLine(title=u"Size", default=u"Quite normal")

class Edit(grok.EditForm):
    pass

class Edit2(grok.EditForm):
    pass

edit2 = grok.PageTemplate('<p>Test edit</p>')

class Edit3(grok.EditForm):
    grok.template('edit2')

class Display(grok.DisplayForm):
    pass

class Display2(grok.DisplayForm):
    pass

display2 = grok.PageTemplate('<p>Test display</p>')

class Display3(grok.DisplayForm):
    grok.template('display2')



"""
Templates that are not associated with a view class will provoke an
error:

  >>> grok.testing.grok(__name__)
  Traceback (most recent call last):
  ...
  ConfigurationExecutionError: martian.error.GrokError: Found the following unassociated template(s) when grokking
  'grok.tests.view.unassociated': index.  Define view classes inheriting
  from grok.View to enable the template(s).
  in:

"""
import grok

class Mammoth(grok.Model):
    pass



"""
  >>> getRootFolder()["manfred"] = Mammoth()

The default view name for a model is 'index':

  >>> from zope.testbrowser.testing import Browser
  >>> browser = Browser()
  >>> browser.handleErrors = False
  >>> browser.open("http://localhost/manfred")
  >>> print browser.contents
  <html>
  <body>
  <h1>Hello, world!</h1>
  <span>Blue</span>
  <span>Blue</span>
  </body>
  </html>

"""
import grok

class Mammoth(grok.Model):
    teeth = u"Blue"

class Index(grok.View):
    pass

index = grok.PageTemplate("""\
<html>
<body>
<h1>Hello, world!</h1>
<span tal:content="python:context.teeth">green</span>
<span tal:content="context/teeth">green</span>
</body>
</html>
""")



"""
A grok.EditForm uses applyData in update mode to save the form data on
the object.  Update mode means that only those fields are changed on
the object that need to be changed.

This is essentially the same narrative as 'editform_applydata'. Here
we test the whole procedure on fields from schemas:

  >>> getRootFolder()["manfred"] = Mammoth('Manfred the Mammoth', 'Really big')

  >>> from zope.testbrowser.testing import Browser
  >>> browser = Browser()
  >>> browser.handleErrors = False

If we don't change any of the fields, there will no object modified
event and applyData will report no changes:

  >>> browser.open("http://localhost/manfred/@@edit")
  >>> browser.getControl("Apply").click()
  >>> 'No changes' in browser.contents
  True

If we change one field, only that attribute will be changed.  The
object modified event also reflects that:

  >>> browser.getControl(name="form.name").value = "Manfred the Big Mammoth"
  >>> browser.getControl("Apply").click()
  The 'name' property is being set.
  An IObjectModifiedEvent was sent for a mammoth with the following changes:
  IMammoth: name
  >>> 'Updated' in browser.contents
  True

Let's change the other field:

  >>> browser.getControl(name="form.size").value = "Enormously big"
  >>> browser.getControl("Apply").click()
  The 'size' property is being set.
  An IObjectModifiedEvent was sent for a mammoth with the following changes:
  IMammoth: size
  >>> 'Updated' in browser.contents
  True

And finally let's change both fields:

  >>> browser.getControl(name="form.name").value = "Manfred the Mammoth"
  >>> browser.getControl(name="form.size").value = "Really big"
  >>> browser.getControl("Apply").click()
  The 'name' property is being set.
  The 'size' property is being set.
  An IObjectModifiedEvent was sent for a mammoth with the following changes:
  IMammoth: name, size
  >>> 'Updated' in browser.contents
  True

"""
import grok
from zope import schema, interface

class IMammoth(interface.Interface):
    name = schema.TextLine(title=u"Name")
    size = schema.TextLine(title=u"Size")

class Mammoth(grok.Model):
    grok.implements(IMammoth)

    def __init__(self, name='', size=''):
        self._name = name
        self._size = size

    def getName(self):
        return self._name
    def setName(self, value):
        print "The 'name' property is being set."
        self._name = value
    name = property(getName, setName)

    def getSize(self):
        return self._size
    def setSize(self, value):
        print "The 'size' property is being set."
        self._size = value
    size = property(getSize, setSize)

class Edit(grok.EditForm):
    form_fields = grok.AutoFields(IMammoth)

@grok.subscribe(Mammoth, grok.IObjectModifiedEvent)
def notify_change_event(mammoth, event):
    print ("An IObjectModifiedEvent was sent for a mammoth with the "
           "following changes:")
    for descr in event.descriptions:
        print descr.interface.__name__ + ": " + ", ".join(descr.attributes)



##############################################################################
#
# Copyright (c) 2007 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Grok publication objects
"""

from grok.rest import GrokMethodNotAllowed

from zope import component
from zope.security.proxy import removeSecurityProxy
from zope.security.checker import selectChecker
from zope.publisher.publish import mapply

from zope.app.publication.http import BaseHTTPPublication, HTTPPublication
from zope.app.publication.browser import BrowserPublication
from zope.app.publication.requestpublicationfactories import \
     BrowserFactory, XMLRPCFactory, HTTPFactory
from zope.app.http.interfaces import IHTTPException

from zope.publisher.interfaces.browser import IBrowserView
from grok.components import View as GrokView
from grok.components import JSON

class ZopePublicationSansProxy(object):

    def getApplication(self, request):
        result = super(ZopePublicationSansProxy, self).getApplication(request)
        return removeSecurityProxy(result)

    def traverseName(self, request, ob, name):
        result = super(ZopePublicationSansProxy, self).traverseName(
            request, ob, name)
        bare_result = removeSecurityProxy(result)
        if IBrowserView.providedBy(bare_result):
            if isinstance(bare_result, (GrokView, JSON)):
                return bare_result
            else:
                return result
        else:
            return bare_result

        
    def callObject(self, request, ob):
        checker = selectChecker(ob)
        if checker is not None:
            checker.check(ob, '__call__')
        return super(ZopePublicationSansProxy, self).callObject(request, ob)


class GrokBrowserPublication(ZopePublicationSansProxy, BrowserPublication):

    def getDefaultTraversal(self, request, ob):
        obj, path = super(GrokBrowserPublication, self).getDefaultTraversal(
            request, ob)
        return removeSecurityProxy(obj), path


class GrokBrowserFactory(BrowserFactory):

    def __call__(self):
        request, publication = super(GrokBrowserFactory, self).__call__()
        return request, GrokBrowserPublication


class GrokXMLRPCPublication(ZopePublicationSansProxy, BaseHTTPPublication):
    pass

class GrokXMLRPCFactory(XMLRPCFactory):

    def __call__(self):
        request, publication = super(GrokXMLRPCFactory, self).__call__()
        return request, GrokXMLRPCPublication


class GrokHTTPPublication(ZopePublicationSansProxy, HTTPPublication):
   def callObject(self, request, ob):
       orig = ob
       if not IHTTPException.providedBy(ob):
           ob = component.queryMultiAdapter((ob, request),
                                            name=request.method)
           checker = selectChecker(ob)
           if checker is not None:
               checker.check(ob, '__call__')
           ob = getattr(ob, request.method, None)
           if ob is None:
               raise GrokMethodNotAllowed(orig, request)
       return mapply(ob, request.getPositionalArguments(), request)

class GrokHTTPFactory(HTTPFactory):
    def __call__(self):
        request, publication = super(GrokHTTPFactory, self).__call__()
        return request, GrokHTTPPublication



"""
  >>> getRootFolder()["manfred"] = Mammoth()

  >>> from zope.testbrowser.testing import Browser
  >>> browser = Browser()
  >>> browser.handleErrors = False

Form variables such as GET parameters are dispatched to arguments of
the render() method, should the method choose to take them:

  >>> browser.open("http://localhost/manfred/render?message=Foo&another=Bar")
  >>> print browser.contents
  Message: Foo
  Another: Bar

Supplying more arguments than those specified has no effect:

  >>> browser.open("http://localhost/manfred/render?message=There&another=Is&last=More")
  >>> print browser.contents
  Message: There
  Another: Is

If you don't supply all of the arguments, there will be a System Error:

  >>> browser.open("http://localhost/manfred/render?message=Foo")
  Traceback (most recent call last):
  ...
  TypeError: Missing argument to render(): another

The same works with views that define update():

  >>> browser.open("http://localhost/manfred/update?message=Foo&another=Bar")
  >>> print browser.contents
  Coming to us from update():
  Message: Foo
  Another: Bar

  >>> browser.open("http://localhost/manfred/update?message=There&another=Is&last=More")
  >>> print browser.contents
  Coming to us from update():
  Message: There
  Another: Is

  >>> browser.open("http://localhost/manfred/update?another=Bar")
  Traceback (most recent call last):
  ...
  TypeError: Missing argument to update(): message

"""
import grok

class Mammoth(grok.Model):
    pass

class RenderWithArguments(grok.View):
    grok.name('render')

    def render(self, message, another):
        return "Message: %s\nAnother: %s" % (message, another)

class UpdateWithArguments(grok.View):
    grok.name('update')
    grok.template('update')

    def update(self, message, another):
        self.message = message
        self.another = another

update = grok.PageTemplate("""
Coming to us from update():
Message: <span tal:replace="view/message" />
Another: <span tal:replace="view/another" />
""")



"""
Anything can be registered as a local utility. If it implements a single
interface, there is no need to specify which interface it provides.

In this test, the utility does not implement any interface, so it cannot be
registered as a local utility.

  >>> grok.testing.grok(__name__)
  Traceback (most recent call last):
    ...
  GrokError: <class 'grok.tests.utility.local_implementsnone.Fireplace'>
  must implement at least one interface (use grok.implements to specify).

"""

import grok
from zope import interface

class Fireplace(object):
    pass

class Cave(grok.Model, grok.Site):
    grok.local_utility(Fireplace)



##############################################################################
#
# Copyright (c) 2006 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Grok
"""
import sys
import types
from zope.component._declaration import adapter as _adapter
from zope.interface.declarations import implementer as _implementer
from martian.util import frame_is_module
from martian.error import GrokImportError

class subscribe:

    def __init__(self, *args):
        self.subscribed = args

    def __call__(self, function):
        frame = sys._getframe(1)
        if not frame_is_module(frame):
            raise GrokImportError("@grok.subscribe can only be used on module "
                                  "level.")

        if not self.subscribed:
            raise GrokImportError("@grok.subscribe requires at least one "
                                  "argument.")

        subscribers = frame.f_locals.get('__grok_subscribers__', None)
        if subscribers is None:
            frame.f_locals['__grok_subscribers__'] = subscribers = []
        subscribers.append((function, self.subscribed))
        return function

class adapter(_adapter):

    def __init__(self, *interfaces):
        # Override the z.c.adapter decorator to force sanity checking
        # and have better error reporting.
        if not interfaces:
            raise GrokImportError(
                "@grok.adapter requires at least one argument.")
        if type(interfaces[0]) is types.FunctionType:
            raise GrokImportError(
                "@grok.adapter requires at least one argument.")
        self.interfaces = interfaces

class implementer(_implementer):

    def __call__(self, ob):
        # XXX we do not have function grokkers (yet) so we put the annotation
        # on the module.
        frame = sys._getframe(1)
        implementers = frame.f_locals.get('__implementers__', None)
        if implementers is None:
            frame.f_locals['__implementers__'] = implementers = []
        implementers.append(ob)
        return _implementer.__call__(self, ob)



"""
Views either need an associated template or a ``render`` method:

  >>> grok.testing.grok(__name__)
  Traceback (most recent call last):
    ...
  ConfigurationExecutionError: martian.error.GrokError: View <class 'grok.tests.view.notemplateorrender.CavePainting'>
  has no associated template or 'render' method.
  in:

"""

import grok

class Mammoth(grok.Model):
    pass

class CavePainting(grok.View):
    pass



from martian.error import GrokError
from martian import util

import os
import zope.component
import grok
import warnings

class TemplateRegistry(object):

    def __init__(self):
        self._reg = {}

    def register(self, name, template):
        self._reg[name] = dict(template=template, associated=False)

    def markAssociated(self, name):
        self._reg[name]['associated'] = True

    def get(self, name):
        entry = self._reg.get(name)
        if entry is None:
            return None
        return entry['template']

    def findFilesystem(self, module_info):
        template_dir_name = module_info.getAnnotation(
            'grok.templatedir', module_info.name + '_templates')

        template_dir = module_info.getResourcePath(template_dir_name)

        if not os.path.isdir(template_dir):
            return
        
        for template_file in os.listdir(template_dir):
            if template_file.startswith('.') or template_file.endswith('~'):
                continue

            template_name, extension = os.path.splitext(template_file)
            extension = extension[1:] # Get rid of the leading dot.
            template_factory = zope.component.queryUtility(
                grok.interfaces.ITemplateFileFactory,
                name=extension)
                
            if template_factory is None:
                # Warning when importing files. This should be
                # allowed because people may be using editors that generate
                # '.bak' files and such.
                warnings.warn("File '%s' has an unrecognized extension in "
                              "directory '%s'" %
                              (template_file, template_dir), UserWarning, 2)
                continue

            inline_template = self.get(template_name)
            if inline_template:
                raise GrokError("Conflicting templates found for name '%s' "
                                "in module %r, either inline and in template "
                                "directory '%s', or two templates with the "
                                "same name and different extensions."
                                % (template_name, module_info.getModule(),
                                   template_dir), inline_template)

            template = template_factory(template_file, template_dir)
            template_path = os.path.join(template_dir, template_file)
            template._annotateGrokInfo(template_name, template_path)

            self.register(template_name, template)

    def listUnassociated(self):
        for name, entry in self._reg.iteritems():
            if not entry['associated']:
                yield name

    def checkUnassociated(self, module_info):
        unassociated = list(self.listUnassociated())
        if unassociated:
            raise GrokError("Found the following unassociated template(s) when "
                            "grokking %r: %s.  Define view classes inheriting "
                            "from grok.View to enable the template(s)."
                            % (module_info.dotted_name,
                               ', '.join(unassociated)), module_info)

    def checkTemplates(self, module_info, factory, factory_name):
        template_name = util.class_annotation(factory, 'grok.template',
                                              factory_name)
            
        if factory_name != template_name:
            # grok.template is being used
    
            if self.get(factory_name):
                raise GrokError("Multiple possible templates for view %r. It "
                                "uses grok.template('%s'), but there is also "
                                "a template called '%s'."
                                % (factory, template_name, factory_name),
                                factory)
        template = self.get(template_name)
        if template:
            if (getattr(factory, 'render', None) and not
                util.check_subclass(factory, grok.components.GrokForm)):
                # we do not accept render and template both for a view
                # (unless it's a form, they happen to have render.
                raise GrokError(
                    "Multiple possible ways to render view %r. "
                    "It has both a 'render' method as well as "
                    "an associated template." % factory, factory)
            self.markAssociated(template_name)
            factory.template = template
            template._initFactory(factory)
        else:
            if not getattr(factory, 'render', None):
                # we do not accept a view without any way to render it
                raise GrokError("View %r has no associated template or "
                                "'render' method." % factory, factory)

class PageTemplateFileFactory(grok.GlobalUtility):
    
    grok.implements(grok.interfaces.ITemplateFileFactory)
    grok.name('pt')
    
    def __call__(self, filename, _prefix=None):
        return grok.components.PageTemplate(filename=filename, _prefix=_prefix)



"""
We can use AddFrom.applyData to save changes to a newly created
object.  The object doesn't yet need to have the attributes that are
going to be set on it.

  >>> getRootFolder()["zoo"] = Zoo()

  >>> from zope.testbrowser.testing import Browser
  >>> browser = Browser()
  >>> browser.handleErrors = False

AddForm.applyData() sends an IObjectModifiedEvent after having
modified the object.  Its return value is True in a Boolean sense when
the object has been modified:

  >>> browser.open("http://localhost/zoo/@@addmammoth")
  >>> browser.getControl(name="form.name").value = "Ellie the Mammoth"
  >>> browser.getControl(name="form.size").value = "Really small"
  >>> browser.getControl("Add entry").click()
  An IObjectModifiedEvent was sent for a mammoth with the following changes:
  IMammoth: name, size
  >>> print browser.contents
  There were changes according to applyData.

  >>> browser.open("http://localhost/zoo/ellie")
  >>> print browser.contents
  Hi, my name is Ellie the Mammoth, and I\'m "Really small"

"""
import grok
from zope import schema, interface

class Zoo(grok.Container):
    pass

class IMammoth(interface.Interface):
    name = schema.TextLine(title=u"Name")
    size = schema.TextLine(title=u"Size")

class Mammoth(grok.Model):
    grok.implements(IMammoth)

class Index(grok.View):
    grok.context(Mammoth)
    def render(self):
        return 'Hi, my name is %s, and I\'m "%s"' % (self.context.name,
                                                     self.context.size)

class AddMammoth(grok.AddForm):
    grok.context(Zoo)

    form_fields = grok.AutoFields(IMammoth)

    @grok.action('Add entry')
    def add(self, **data):
        self.context['ellie'] = ellie = Mammoth()
        if self.applyData(ellie, **data):
            return 'There were changes according to applyData.'
        return 'There were no changes according to applyData.'

@grok.subscribe(Mammoth, grok.IObjectModifiedEvent)
def notify_change_event(mammoth, event):
    print ("An IObjectModifiedEvent was sent for a mammoth with the "
           "following changes:")
    for descr in event.descriptions:
        print descr.interface.__name__ + ": " + ", ".join(descr.attributes)



"""
A form view can completely override which fields are displayed by setting
form_fields manually:

  >>> grok.testing.grok(__name__)

We only expect a single field to be present in the form, as we omitted 'size':

  >>> from zope import component
  >>> from zope.publisher.browser import TestRequest
  >>> request = TestRequest()
  >>> view = component.getMultiAdapter((Mammoth(), request), name='edit')
  >>> len(view.form_fields)
  1
  >>> [w.__name__ for w in view.form_fields]
  ['name']

"""

import grok
from zope import interface, schema

class IMammoth(interface.Interface):
    name = schema.TextLine(title=u"Name")
    size = schema.TextLine(title=u"Size", default=u"Quite normal")

class Mammoth(grok.Model):
    interface.implements(IMammoth)

class Edit(grok.EditForm):
    grok.context(Mammoth)

    form_fields = grok.Fields(IMammoth).omit('size')



"""
A grok.Model may implement one or more interfaces that are schemas:

  >>> grok.testing.grok(__name__)
  >>> manfred = Mammoth()
  >>> print manfred.name
  None
  >>> print manfred.size
  Quite normal

A grok.EditForm is a special grok.View that renders an edit form.

  >>> from zope import component
  >>> from zope.publisher.browser import TestRequest
  >>> request = TestRequest()
  >>> view = component.getMultiAdapter((manfred, request), name='edit')
  >>> len(view.form_fields)
  2
  >>> [w.__name__ for w in view.form_fields]
  ['name', 'size']

When there are multiple schemas in play, we get all the fields:

  >>> view = component.getMultiAdapter((Manfred(), request), name='edit2')
  >>> len(view.form_fields)
  3
  >>> [w.__name__ for w in view.form_fields]
  ['can_talk', 'name', 'size']

If the context is an interface instead of a model directly, the fields
will be retrieved from that interface, and that interface only:

  >>> view = component.getMultiAdapter(
  ...   (YetAnotherMammoth(), request), name='edit4')
  >>> len(view.form_fields)
  2
  >>> [w.__name__ for w in view.form_fields]
  ['alpha', 'beta']

"""
import grok
from zope import interface, schema
from zope.schema.fieldproperty import FieldProperty

class IMammoth(interface.Interface):
    name = schema.TextLine(title=u"Name")
    size = schema.TextLine(title=u"Size", default=u"Quite normal")


class Mammoth(grok.Model):
    interface.implements(IMammoth)

    name = FieldProperty(IMammoth['name'])    
    size = FieldProperty(IMammoth['size'])


class Edit(grok.EditForm):
    grok.context(Mammoth)


class IMovieCharacter(interface.Interface):
    can_talk = schema.Bool(title=u'Can talk', default=False)


class Manfred(Mammoth):
    interface.implements(IMovieCharacter)


class Edit2(grok.EditForm):
    grok.context(Manfred)
    

class IYetAnotherMammoth(interface.Interface):
    alpha = schema.TextLine(title=u'alpha')
    beta = schema.TextLine(title=u'beta')


class YetAnotherMammoth(grok.Model):
    interface.implements(IYetAnotherMammoth)


class Edit4(grok.EditForm):
    grok.context(IYetAnotherMammoth)



"""
When a view's update() method redirects somewhere else, the template
is not executed subsequently.

  >>> grok.testing.grok(__name__)

  >>> manfred = Mammoth()
  >>> from zope.publisher.browser import TestRequest
  >>> request = TestRequest()
  >>> from zope.component import getMultiAdapter
  >>> view = getMultiAdapter((manfred, request), name='cavepainting')
  >>> print view()
  None
  >>> print request.response.getStatus()
  302
  >>> print request.response.getHeader('Location')
  somewhere-else

"""
import grok

class Mammoth(grok.Model):
    pass

class CavePainting(grok.View):
    def update(self):
        self.request.response.redirect('somewhere-else')


cavepainting = grok.PageTemplate("""\
<html>
<body>
<h1 tal:content="this-is-an-error" />
</body>
</html>
""")



"""
Templates with ambiguous context cannot be grokked:

  >>> grok.testing.grok(__name__)
  Traceback (most recent call last):
    ...
  GrokError: Multiple possible contexts for
  <class 'grok.tests.view.ambiguouscontext.Club'>, please use grok.context.

"""

import grok

class Cave(grok.Model):
    pass

class Mammoth(grok.Model):
    pass

class Club(grok.View):
    pass



##############################################################################
#
# Copyright (c) 2007 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""

The admin-UI now supports detection and deletion of broken objects. It
is still limited to IApplication objects in the root folder. Options
to repair broken objects are also still missing.

We first setup the environment:

  >>> from zope.testbrowser.testing import Browser
  >>> browser = Browser()
  >>> browser.addHeader('Authorization', 'Basic mgr:mgrpw')

If no broken applications are in the root, everything should look as
usual:

  >>> browser.open('http://localhost/applications')
  >>> 'Broken applications:' not in browser.contents
  True

Now we grok this module, to have a new application type available,
which is intentionally broken:

  >>> browser.open('http://localhost/applications')
  >>> 'PseudoBroken' in browser.contents
  True

We add an instance of that new type:

  >>> subform = browser.getForm(name='PseudoBroken')
  >>> subform
  <zope.testbrowser.browser.Form object at 0x...>

  >>> subform.getControl('Name your new app').value = 'mybrokenobj'
  >>> subform.getControl('Create').click()

and the broken object should show up in the applications list:

 >>> print browser.contents
 <html xmlns="http://www.w3.org/1999/xhtml">
 ...
 ...Currently no working applications are...installed...
 ...
 ...Broken applications:...
 ...
 ...(broken type: grok.ftests.admin.brokenobjs.PseudoBroken)...
 ...This application is broken!...
 ...

If we want to delete the broken object, we can do so:

  >>> ctrl = browser.getControl(name='items')
  >>> ctrl.getControl(value='mybrokenobj').selected = True
  >>> browser.getControl('Delete Selected').click()
  >>> print browser.contents
  <html xmlns="http://www.w3.org/1999/xhtml">
  ...
  ...Application `mybrokenobj` was successfully deleted.
  ...

and the 'Broken applications' section won't show up anymore:

  >>> 'Broken applications:' not in browser.contents
  True


"""
import grok
from ZODB.broken import Broken
class PseudoBroken(grok.Application, grok.Container, Broken):
    """A class intentionally broken.
    """
    pass



"""
A base class of something that can be a context (such as a model) can
function as a module-level context, and thus can have views associated
with it.

  >>> grok.testing.grok(__name__)

  >>> from zope.publisher.browser import TestRequest
  >>> request = TestRequest()
  >>> from zope import component
  >>> model = ModelBase()
  >>> view = component.getMultiAdapter((model, request), name='viewbase')
  Traceback (most recent call last):
    ...
  ComponentLookupError: ((<grok.tests.baseclass.basecontext.ModelBase object at 0x...>,
  <zope.publisher.browser.TestRequest instance ...>),
  <InterfaceClass zope.interface.Interface>,
  'viewbase')
  >>> view = component.getMultiAdapter((model, request), name='realview')
  >>> view.render()
  'hello world'
"""

import grok

class ModelBase(grok.Model):
    pass

class ViewBase(grok.View):    
    def render(self):
        return "hello world"

class RealView(ViewBase):
    pass



"""
  >>> getRootFolder()["manfred"] = Mammoth()

  >>> from zope.testbrowser.testing import Browser
  >>> browser = Browser()
  >>> browser.handleErrors = False
  >>> browser.open("http://localhost/manfred/@@painting")
  >>> print browser.contents
  <html>
  <body>
  <h1>GROK MACRO!</h1>
  <div>
  GROK SLOT!
  </div>
  </body>
  </html>

Views without a template do not support macros:

  >>> browser.open("http://localhost/manfred/@@dancing")
  Traceback (most recent call last):
  AttributeError: 'DancingHall' object has no attribute 'template'

If the view has an attribute with the same name as a macro, the macro 
shadows the view. XXX This should probably generate a warning at runtime.

  >>> browser.open("http://localhost/manfred/@@grilldish")
  >>> print browser.contents
  <html>
  Curry
  </html>
  
You can skip the "macro" part of the macro call, but this is deprecated:

  >>> from zope.deprecation.tests import warn
  >>> import warnings
  >>> saved_warn = warnings.warn
  >>> warnings.warn = warn

  >>> browser.open("http://localhost/manfred/@@burnt")
  From tests.py's showwarning():
  ... DeprecationWarning: Calling macros directly on the view is deprecated. Please use context/@@viewname/macros/macroname
  ...
  
  >>> warnings.warn = saved_warn

"""
import grok

class Mammoth(grok.Model):
    pass

class DancingHall(grok.View):

    def render(self):
        return "A nice large dancing hall for mammoths."

class Grilled(grok.View):

    def update(self):
        self.spices = "Pepper and salt"

class Painting(grok.View):
    pass

painting = grok.PageTemplate("""\
<html metal:use-macro="context/@@layout/macros/main">
<div metal:fill-slot="slot">
GROK SLOT!
</div>
</html>
""")

class Layout(grok.View):
    pass

layout = grok.PageTemplate("""\
<html metal:define-macro="main">
<body>
<h1>GROK MACRO!</h1>
<div metal:define-slot="slot">
</div>
</body>
</html>""")

class Dancing(grok.View):
    pass

dancing = grok.PageTemplate("""\
<html metal:use-macro="context/@@dancinghall/macros/something">
</html>
""")

class GrillDish(grok.View):
    pass

grilldish = grok.PageTemplate("""
<html metal:use-macro="context/@@grilled/macros/spices">
</html>""")

class Burnt(grok.View):
    pass

burnt = grok.PageTemplate("""\
<html metal:use-macro="context/@@grilled/spices">
</html>""")

class Grilled(grok.View):
    pass

grilled = grok.PageTemplate("""\
<html metal:define-macro="spices">
Curry
</html>""")




##############################################################################
#
# Copyright (c) 2006-2007 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Grok components"""

import sys
import os
import persistent
import urllib
import datetime
import warnings
import pytz
import simplejson

from zope import component
from zope import interface
from zope.interface.common import idatetime
from zope.security.permission import Permission
from zope.app.securitypolicy.role import Role
from zope.publisher.browser import BrowserPage
from zope.publisher.interfaces import NotFound
from zope.publisher.interfaces.browser import (IBrowserPublisher,
                                               IBrowserRequest)
from zope.publisher.publish import mapply
from zope.pagetemplate import pagetemplate, pagetemplatefile
from zope.formlib import form
from zope.annotation.interfaces import IAttributeAnnotatable

from zope.app.pagetemplate.engine import TrustedAppPT
from zope.app.publisher.browser import getDefaultViewName
from zope.app.publisher.browser import directoryresource
from zope.app.publisher.browser.pagetemplateresource import \
    PageTemplateResourceFactory
from zope.app.container.btree import BTreeContainer
from zope.app.container.contained import Contained
from zope.app.container.interfaces import IReadContainer, IObjectAddedEvent
from zope.app.component.site import SiteManagerContainer
from zope.app.component.site import LocalSiteManager

import z3c.flashmessage.interfaces

import martian.util
from grok import interfaces, formlib, util


class Model(Contained, persistent.Persistent):
    # XXX Inheritance order is important here. If we reverse this,
    # then containers can't be models anymore because no unambigous MRO
    # can be established.
    interface.implements(IAttributeAnnotatable)


class Container(BTreeContainer):
    interface.implements(IAttributeAnnotatable)


class Site(SiteManagerContainer):
    pass

@component.adapter(Site, IObjectAddedEvent)
def addSiteHandler(site, event):
    sitemanager = LocalSiteManager(site)
    # LocalSiteManager creates the 'default' folder in its __init__.
    # It's not needed anymore in new versions of Zope 3, therefore we
    # remove it
    del sitemanager['default']
    site.setSiteManager(sitemanager)


class Application(Site):
    """A top-level application object."""
    interface.implements(interfaces.IApplication)


class Adapter(object):

    def __init__(self, context):
        self.context = context


class GlobalUtility(object):
    pass


class LocalUtility(Model):
    pass


class MultiAdapter(object):
    pass


class Annotation(persistent.Persistent):
    pass


class ViewBase(object):
    def __init__(self, context, request):
        self.context = context
        self.request = request

class View(BrowserPage):
    interface.implements(interfaces.IGrokView)

    def __init__(self, context, request):
        super(View, self).__init__(context, request)
        self.__name__ = self.__view_name__
        self.static = component.queryAdapter(
            self.request,
            interface.Interface,
            name=self.module_info.package_dotted_name
            )

    @property
    def response(self):
        return self.request.response

    def __call__(self):
        mapply(self.update, (), self.request)
        if self.request.response.getStatus() in (302, 303):
            # A redirect was triggered somewhere in update().  Don't
            # continue rendering the template or doing anything else.
            return

        template = getattr(self, 'template', None)
        if template is not None:
            return self._render_template()
        return mapply(self.render, (), self.request)

    def _render_template(self):
        return self.template.render(self)

    def namespace(self):
        return {}

    def __getitem__(self, key):
        # This is BBB code for Zope page templates only:
        if not isinstance(self.template, PageTemplate):
            raise AttributeError("View has no item %s" % key)
        
        value = self.template._template.macros[key]
        # When this deprecation is done with, this whole __getitem__ can 
        # be removed.
        warnings.warn("Calling macros directly on the view is deprecated. "
                      "Please use context/@@viewname/macros/macroname\n"
                      "View %r, macro %s" % (self, key),
                      DeprecationWarning, 1)
        return value

    
    def url(self, obj=None, name=None):
        # if the first argument is a string, that's the name. There should
        # be no second argument
        if isinstance(obj, basestring):
            if name is not None:
                raise TypeError(
                    'url() takes either obj argument, obj, string arguments, '
                    'or string argument')
            name = obj
            obj = None

        if name is None and obj is None:
            # create URL to view itself
            obj = self
        elif name is not None and obj is None:
            # create URL to view on context
            obj = self.context
        return util.url(self.request, obj, name)

    def application_url(self, name=None):
        obj = self.context
        while obj is not None:
            if isinstance(obj, Application):
                return self.url(obj, name)
            obj = obj.__parent__
        raise ValueError("No application found.")

    def redirect(self, url):
        return self.request.response.redirect(url)

    def update(self):
        pass

    def flash(self, message, type='message'):
        source = component.getUtility(
            z3c.flashmessage.interfaces.IMessageSource, name='session')
        source.send(message, type)


class XMLRPC(object):
    pass

class REST(object):
    interface.implements(interfaces.IREST)
    
    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.body = request.bodyStream.getCacheStream().read()

    @property
    def response(self):
        return self.request.response
    
##     def GET(self):
##         raise GrokMethodNotAllowed(self.context, self.request)
    
##     def POST(self):
##         raise GrokMethodNotAllowed(self.context, self.request)
    
##     def PUT(self):
##         raise GrokMethodNotAllowed(self.context, self.request)
    
##     def DELETE(self):
##         raise GrokMethodNotAllowed(self.context, self.request)

class JSON(BrowserPage):

    def __call__(self):
        view_name = self.__view_name__
        method = getattr(self, view_name)
        method_result = mapply(method, (), self.request)
        return simplejson.dumps(method_result)


class BaseTemplate(object):
    """Any sort of page template"""
    
    interface.implements(interfaces.ITemplate)

    __grok_name__ = ''
    __grok_location__ = ''

    def __repr__(self):
        return '<%s template in %s>' % (self.__grok_name__,
                                        self.__grok_location__)

    def _annotateGrokInfo(self, name, location):
        self.__grok_name__ = name
        self.__grok_location__ = location

    def _initFactory(self, factory):
        pass


class GrokTemplate(BaseTemplate):
    """A slightly more advanced page template
    
    This provides most of what a page template needs and is a good base for
    writing your own page template"""
    
    def __init__(self, string=None, filename=None, _prefix=None):

        # __grok_module__ is needed to make defined_locally() return True for
        # inline templates
        # XXX unfortunately using caller_module means that care must be taken
        # when GrokTemplate is subclassed. You can not do a super().__init__
        # for example.
        self.__grok_module__ = martian.util.caller_module()
        
        if not (string is None) ^ (filename is None):
            raise AssertionError("You must pass in template or filename, but not both.")
        
        if string:
            self.setFromString(string)
        else:
            if _prefix is None:
                module = sys.modules[self.__grok_module__]
                _prefix = os.path.dirname(module.__file__)
            self.setFromFilename(filename, _prefix)

    def __repr__(self):
        return '<%s template in %s>' % (self.__grok_name__,
                                        self.__grok_location__)
    
    def _annotateGrokInfo(self, name, location):
        self.__grok_name__ = name
        self.__grok_location__ = location

    def _initFactory(self, factory):
        pass

    def namespace(self, view):
        namespace = {}
        namespace['request'] = view.request
        namespace['view'] = view
        namespace['context'] = view.context
        # XXX need to check whether we really want to put None here if missing
        namespace['static'] = view.static
        
        return namespace
    
    def getNamespace(self, view):
        namespace = self.namespace(view)
        namespace.update(view.namespace())
        return namespace

class TrustedPageTemplate(TrustedAppPT, pagetemplate.PageTemplate):
    pass

class TrustedFilePageTemplate(TrustedAppPT, pagetemplatefile.PageTemplateFile):
    pass

class PageTemplate(GrokTemplate):
    
    def setFromString(self, string):
        zpt = TrustedPageTemplate()
        if martian.util.not_unicode_or_ascii(string):
            raise ValueError("Invalid page template. Page templates must be "
                             "unicode or ASCII.")
        zpt.write(string)
        self._template = zpt

    def setFromFilename(self, filename, _prefix=None):
        self._template = TrustedFilePageTemplate(filename, _prefix)

    def _initFactory(self, factory):
        factory.macros = self._template.macros

    def render(self, view):
        namespace = self.getNamespace(view)
        template = self._template
        namespace.update(template.pt_getContext())
        return template.pt_render(namespace)
    
class PageTemplateFile(PageTemplate):
    # For BBB
    def __init__(self, filename, _prefix=None):
        self.__grok_module__ = martian.util.caller_module()
        if _prefix is None:
            module = sys.modules[self.__grok_module__]
            _prefix = os.path.dirname(module.__file__)
        self.setFromFilename(filename, _prefix)
    
class DirectoryResource(directoryresource.DirectoryResource):
    # We subclass this, because we want to override the default factories for
    # the resources so that .pt and .html do not get created as page
    # templates

    resource_factories = {}
    for type, factory in (directoryresource.DirectoryResource.
                          resource_factories.items()):
        if factory is PageTemplateResourceFactory:
            continue
        resource_factories[type] = factory

class DirectoryResourceFactory(directoryresource.DirectoryResourceFactory):
    # We need this to allow hooking up our own GrokDirectoryResource
    # and to set the checker to None (until we have our own checker)

    def __call__(self, request):
        # Override this method for the following line, in which our
        # custom DirectoryResource class is instantiated.
        resource = DirectoryResource(self.__dir, request)
        resource.directory_factory = DirectoryResourceFactory
        resource.__Security_checker__ = self.__checker
        resource.__name__ = self.__name
        return resource

class Traverser(object):
    interface.implements(IBrowserPublisher)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def browserDefault(self, request):
        # if we have a RESTful request, we will handle
        # GET, POST and HEAD differently (PUT and DELETE are handled already
        # but not on the BrowserRequest layer but the HTTPRequest layer)
        if IRESTLayer.providedBy(request):
            rest_view = component.getMultiAdapter(
                (self.context, self.request),
                name=request.method)
            return rest_view, ()
        view_name = getDefaultViewName(self.context, request)
        view_uri = "@@%s" % view_name
        return self.context, (view_uri,)

    def publishTraverse(self, request, name):
        subob = self.traverse(name)
        if subob is not None:
            return util.safely_locate_maybe(subob, self.context, name)

        # XXX Special logic here to deal with containers.  It would be
        # good if we wouldn't have to do this here. One solution is to
        # rip this out and make you subclass ContainerTraverser if you
        # wanted to override the traversal behaviour of containers.
        if IReadContainer.providedBy(self.context):
            item = self.context.get(name)
            if item is not None:
                return item

        view = component.queryMultiAdapter((self.context, request), name=name)
        if view is not None:
            return view

        raise NotFound(self.context, name, request)

    def traverse(self, name):
        # this will be overridden by subclasses
        pass


class ModelTraverser(Traverser):
    component.adapts(Model, IBrowserRequest)

    def traverse(self, name):
        traverse = getattr(self.context, 'traverse', None)
        if traverse:
            return traverse(name)


class ContainerTraverser(Traverser):
    component.adapts(Container, IBrowserRequest)

    def traverse(self, name):
        traverse = getattr(self.context, 'traverse', None)
        if traverse:
            result = traverse(name)
            if result is not None:
                return result
        # try to get the item from the container
        return self.context.get(name)


default_form_template = PageTemplateFile(os.path.join(
    'templates', 'default_edit_form.pt'))
default_form_template.__grok_name__ = 'default_edit_form'
default_display_template = PageTemplateFile(os.path.join(
    'templates', 'default_display_form.pt'))
default_display_template.__grok_name__ = 'default_display_form'


class GrokForm(object):
    """Mix-in to console zope.formlib's forms with grok.View and to
    add some more useful methods.

    The consolation needs to happen because zope.formlib's Forms have
    update/render methods which have different meanings than
    grok.View's update/render methods.  We deal with this issue by
    'renaming' zope.formlib's update() to update_form() and by
    disallowing subclasses to have custom render() methods."""

    def update(self):
        """Subclasses can override this method just like on regular
        grok.Views. It will be called before any form processing
        happens."""

    def update_form(self):
        """Update the form, i.e. process form input using widgets.

        On zope.formlib forms, this is what the update() method is.
        In grok views, the update() method has a different meaning.
        That's why this method is called update_form() in grok forms."""
        super(GrokForm, self).update()

    def render(self):
        """Render the form, either using the form template or whatever
        the actions returned in form_result."""
        # if the form has been updated, it will already have a result
        if self.form_result is None:
            if self.form_reset:
                # we reset, in case data has changed in a way that
                # causes the widgets to have different data
                self.resetForm()
                self.form_reset = False
            self.form_result = self._render_template()

        return self.form_result

    # Mark the render() method as a method from the base class. That
    # way we can detect whether somebody overrides render() in a
    # subclass (which we don't allow).
    render.base_method = True

    def __call__(self):
        mapply(self.update, (), self.request)
        if self.request.response.getStatus() in (302, 303):
            # A redirect was triggered somewhere in update().  Don't
            # continue rendering the template or doing anything else.
            return

        self.update_form()
        return self.render()


class Form(GrokForm, form.FormBase, View):
    # We're only reusing the form implementation from zope.formlib, we
    # explicitly don't want to inherit the interface semantics (mostly
    # for the different meanings of update/render).
    interface.implementsOnly(interfaces.IGrokForm)

    template = default_form_template

    def applyData(self, obj, **data):
        return formlib.apply_data_event(obj, self.form_fields, data,
                                        self.adapters)

    # BBB -- to be removed in June 2007
    def applyChanges(self, obj, **data):
        warnings.warn("The 'applyChanges' method on forms is deprecated "
                      "and will disappear by June 2007. Please use "
                      "'applyData' instead.", DeprecationWarning, 2)
        return bool(self.applyData(obj, **data))


class AddForm(Form):
    pass


class EditForm(GrokForm, form.EditFormBase, View):
    # We're only reusing the form implementation from zope.formlib, we
    # explicitly don't want to inherit the interface semantics (mostly
    # for the different meanings of update/render).
    interface.implementsOnly(interfaces.IGrokForm)

    template = default_form_template

    def applyData(self, obj, **data):
        return formlib.apply_data_event(obj, self.form_fields, data,
                                        self.adapters, update=True)

    # BBB -- to be removed in June 2007
    def applyChanges(self, obj, **data):
        warnings.warn("The 'applyChanges' method on forms is deprecated "
                      "and will disappear by June 2007. Please use "
                      "'applyData' instead.", DeprecationWarning, 2)
        return bool(self.applyData(obj, **data))

    @formlib.action("Apply")
    def handle_edit_action(self, **data):
        if self.applyData(self.context, **data):
            formatter = self.request.locale.dates.getFormatter(
                'dateTime', 'medium')

            try:
                time_zone = idatetime.ITZInfo(self.request)
            except TypeError:
                time_zone = pytz.UTC

            self.status = "Updated on %s" % formatter.format(
                datetime.datetime.now(time_zone)
                )
        else:
            self.status = 'No changes'


class DisplayForm(GrokForm, form.DisplayFormBase, View):
    # We're only reusing the form implementation from zope.formlib, we
    # explicitly don't want to inherit the interface semantics (mostly
    # for the different meanings of update/render).
    interface.implementsOnly(interfaces.IGrokForm)

    template = default_display_template


class IndexesClass(object):
    def __init__(self, name, bases=(), attrs=None):
        if attrs is None:
            return
        # make sure we take over a bunch of possible attributes
        for name in ['__grok_context__', '__grok_name__',
                     '__grok_site__']:
            value = attrs.get(name)
            if value is not None:
                setattr(self, name, value)
        # now read and store indexes
        indexes = {}
        for name, value in attrs.items():
            if not interfaces.IIndexDefinition.providedBy(value):
                continue
            indexes[name] = value
        self.__grok_indexes__ = indexes
        # __grok_module__ is needed to make defined_locally() return True for
        # inline templates
        self.__grok_module__ = martian.util.caller_module()

Indexes = IndexesClass('Indexes')

class Permission(Permission):
    pass

class Role(Role):
    pass

class IGrokLayer(interface.Interface):
    pass

class IRESTLayer(interface.Interface):
    pass

class Skin(object):
    pass

class RESTProtocol(object):
    pass



"""

  >>> grok.testing.grok(__name__)

  >>> cave = Cave()
  >>> from zope.publisher.browser import TestRequest
  >>> request = TestRequest()
  >>> from zope.component import getMultiAdapter

  >>> view = getMultiAdapter((cave, request), name='piepmatz')
  >>> print view()
  <p>Piep! Piep!</p>

"""

import grok

class Cave(grok.Model):
    pass

class Piepmatz(grok.View):
    pass # template in zpt_templates/piepmatz.pt



##############################################################################
#
# Copyright (c) 2007 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Information about objects.

This module provides wrappers/proies to give information about
objects at runtime. See inspect.txt to learn more about retrieving
object information using ObjectInfos.
"""
import inspect
import re
import types
from types import MethodType

import grok
from grok.admin.utilities import isContainingEvilRegExpChars

import zope
from zope.interface import Interface, implementedBy
from zope.interface.common.sequence import IExtendedReadSequence
from zope.interface.common.mapping import IEnumerableMapping
from zope.annotation.interfaces import IAnnotatable, IAnnotations
from zope.app.apidoc import utilities
from zope.location import location
from zope.schema import getFields
from zope.security.checker import getCheckerForInstancesOf
from zope.security.proxy import removeSecurityProxy
from zope.traversing.api import getParent, getRoot
from zope.traversing.interfaces import IPhysicallyLocatable
from zope.dublincore.interfaces import IZopeDublinCore
from zope.app.folder import rootFolder



class ObjectInfo(grok.Model):
    """An inspect proxy to provide object specific data.

    This is a base proxy, which can handle all (and only)
    Python-related data. See objectinfo.txt to learn more about
    ObjectInfos.
    """
    obj = None

    def __init__(self, obj):
        self.obj = obj

    def getmembers(self, predicate=None):
        """Return all the members of an object in a list of (name,
        value) pairs sorted by name.

        If the optional predicate argument is supplied, only members
        for which the predicate returns a true value are included.
        """
        return inspect.getmembers(self.obj, predicate)

    def getmoduleinfo(self):
        """Get information about modules.

        Return a tuple of values that describe how Python will
        interpret the file identified by path if it is a module, or
        None if it would not be identified as a module. The return
        tuple is (name, suffix, mode, mtype), where name is the name
        of the module without the name of any enclosing package,
        suffix is the trailing part of the file name (which may not be
        a dot-delimited extension), mode is the open() mode that would
        be used ('r' or 'rb'), and mtype is an integer giving the type
        of the module. mtype will have a value which can be compared
        to the constants defined in the imp module; see the
        documentation for that module for more information on module
        types.
        """
        path = getattr(self.obj, '__file__', None)
        if path is None:
            return
        return inspect.getmoduleinfo(path)

    def getmodulename(self):
        """Return the name of the module named by the file path,
        without including the names of enclosing packages. This uses
        the same algorithm as the interpreter uses when searching for
        modules. If the name cannot be matched according to the
        interpreter's rules, None is returned.
        """
        path = getattr(self.obj, '__file__', None)
        if path is None:
            return
        return inspect.getmodulename(path)

    def ismodule(self):
        """Return true if the object is a module.
        """
        return inspect.ismodule(self.obj)

    def isclass(self):
        """Return true if the object is a class.
        """
        return inspect.isclass(self.obj)

    def ismethod(self):
        """Return true if the object is a method.
        """
        return inspect.ismethod(self.obj)

    def isfunction(self):
        """Return true if the object is a Python function or unnamed
        (lambda) function.
        """
        return inspect.isfunction(self.obj)

    def istraceback(self):
        """Return true if the object is a traceback.
        """
        return inspect.istraceback(self.obj)

    def isframe(self):
        """Return true if the object is a frame.
        """
        return inspect.isframe(self.obj)

    def iscode(self):
        """Return true if the object is a code.
        """
        return inspect.iscode(self.obj)

    def isbuiltin(self):
        """Return true if the object is a built-in function.
        """
        return inspect.isbuiltin(self.obj)

    def isroutine(self):
        """Return true if the object is a user-defined or built-in
        function or method.
        """
        return inspect.isroutine(self.obj)

    def ismethoddescriptor(self):
        """Return true if the object is a method descriptor, but not
        if ismethod() or isclass() or isfunction() are true.

        This is new as of Python 2.2, and, for example, is true of
        int.__add__. An object passing this test has a __get__
        attribute but not a __set__ attribute, but beyond that the set
        of attributes varies. __name__ is usually sensible, and
        __doc__ often is.

        Methods implemented via descriptors that also pass one of the
        other tests return false from the ismethoddescriptor() test,
        simply because the other tests promise more - you can, e.g.,
        count on having the im_func attribute (etc) when an object
        passes ismethod().
        """
        return inspect.ismethoddescriptor(self.obj)

    def isdatadescriptor(self):
        """Return true if the object is a data descriptor.

        Data descriptors have both a __get__ and a __set__
        attribute. Examples are properties (defined in Python),
        getsets, and members. The latter two are defined in C and
        there are more specific tests available for those types, which
        is robust across Python implementations. Typically, data
        descriptors will also have __name__ and __doc__ attributes
        (properties, getsets, and members have both of these
        attributes), but this is not guaranteed. New in version 2.3.
        """
        return inspect.isdatadescriptor(self.obj)

    def isgetsetdescriptor(self):
        """Return true if the object is a getset descriptor.

        getsets are attributes defined in extension modules via
        PyGetSetDef structures. For Python implementations without
        such types, this method will always return False.

        New in version 2.5.
        """
        return inspect.isgetsetdescriptor(self.obj)

    def ismemberdescriptor(self):
        """Return true if the object is a member descriptor.

        Member descriptors are attributes defined in extension modules
        via PyMemberDef structures. For Python implementations without
        such types, this method will always return False.

        New in version 2.5.
        """
        return inspect.ismemberdescriptor(self.obj)

    #
    # Source code related...
    #
    def getdoc(self):
        """Get the documentation string for an object.

        All tabs are expanded to spaces. To clean up docstrings that
        are indented to line up with blocks of code, any whitespace
        than can be uniformly removed from the second line onwards is
        removed.
        """
        return inspect.getdoc(self.obj)

    def getcomments(self):
        """Get comments for an object.

        Return in a single string any lines of comments immediately
        preceding the object's source code (for a class, function, or
        method), or at the top of the Python source file (if the
        object is a module).

        Due to a bug in ``inspect.getsource()`` no objects can be
        handled, which contain a regular expression specific char in
        their string representation.

        """
        if isContainingEvilRegExpChars(str(self.obj)):
            return
        return inspect.getcomments(self.obj)

    def getfile(self):
        """Return the name of the (text or binary) file in which an
        object was defined.

        If the object is a built-in module, class or function,
        ``None`` will be returned.
        """
        try:
            return inspect.getfile(self.obj)
        except TypeError:
            return

    def getmodule(self):
        """Try to guess which module an object was defined in.
        """
        return inspect.getmodule(self.obj)

    def getsourcefile(self):
        """Return the name of the Python source file in which an
        object was defined.

        If the object is a built-in module, class or function,
        ``None`` will be returned.
        """
        try:
            return inspect.getsourcefile(self.obj)
        except TypeError:
            return

    def getsourcelines(self):
        """Return a list of source lines and starting line number for
        an object.

        The argument may be a module, class, method, function,
        traceback, frame, or code object. The source code is returned
        as a list of the lines corresponding to the object and the
        line number indicates where in the original source file the
        first line of code was found. An IOError is raised if the
        source code cannot be retrieved.
        """
        try:
            return inspect.getsourcelines(self.obj)
        except TypeError:
            return
        return 

    def getsource(self):
        """Return the text of the source code for an object.

        The argument may be a module, class, method, function,
        traceback, frame, or code object. The source code is returned
        as a single string. An IOError is raised if the source code
        cannot be retrieved.

        Due to a bug in ``inspect.getsource()`` no objects can be
        handled, which contain a regular expression specific char in
        their string representation.
        """
        if isContainingEvilRegExpChars(str(self.obj)):
            return

        try:
            return inspect.getsource(self.obj)
        except TypeError:
            return
        return

    def traverse(self, name):
        new_obj = None

        # Try to get name as dict entry...
        keygetter = getattr(self.obj, 'keys', None)
        if inspect.ismethod(keygetter):
            if name in keygetter():
                new_obj = self.obj[name]

        # Try to get name as sequence entry...
        if not new_obj:
            # This is not the appropriate way to handle iterators. We
            # must find somehing to handle them too.
            try:
                name_int = int(name)
                if name_int in range(0, len(self.obj)):
                    new_obj = self.obj[name_int]
            except ValueError:
                pass

        # Get name as obj attribute...
        if not new_obj and hasattr(self.obj, name):
            new_obj = getattr(self.obj, name, None)

        # Get name as annotation...
        if not new_obj:
            naked = zope.security.proxy.removeSecurityProxy(self.obj)
            try:
                annotations = IAnnotations(naked)
                new_obj = name and name in annotations and annotations[name]
                if not new_obj:
                    new_obj = annotations
            except TypeError:
                pass

        # Give obj a location...
        if new_obj:
            if not IPhysicallyLocatable(new_obj, False):
                new_obj = location.LocationProxy(
                    new_obj, self.obj, name)

            new_info = ZopeObjectInfo(new_obj)
            new_info.__parent__ = self
            new_info.__name__ = name
            return new_info

        # name not found...
        return





class ZopeObjectInfo(ObjectInfo):
    """Zope specific data.
    """

    def __init__(self, obj):
        self.obj = obj
        self.__klass = getattr(obj, '__class__', None) or type(obj)
        self.__interfaces = tuple(implementedBy(self.__klass))

    def getTypeLink(self, obj_type):
        if obj_type is types.NoneType:
            return
        path = utilities.getPythonPath(obj_type)
        return path.replace('.', '/')

    def isLinkable(self, obj):
        """We consider all but some basic types to be linkable for docgrok.

        Although even simple strings can be displayed by a docgrok, it
        does not make much sense. We therefore simply forbid such
        links, filtering objects of basic types out.
        """
        for typespec in [types.NoneType, types.TypeType, types.BooleanType,
                         types.IntType, types.LongType, types.FloatType,
                         types.ComplexType, types.StringTypes,
                         types.MethodType, types.BuiltinFunctionType,
                         types.LambdaType, types.GeneratorType, types.CodeType,
                         types.FileType, types.TracebackType, types.FrameType,
                         types.BufferType, types.NotImplementedType]:
            if isinstance(obj, typespec):
                return False
        return True

    def getParent(self):
        return getParent(self.obj)

    def getPythonPath(self):
        return utilities.getPythonPath(self.obj)

    def getDirectlyProvidedInterfaces(self):
        # This comes from apidoc...
        # Getting the directly provided interfaces works only on naked objects
        naked = removeSecurityProxy(self.obj)
        return [utilities.getPythonPath(iface)
                for iface in zope.interface.directlyProvidedBy(naked)]

    def getProvidedInterfaces(self):
        return self.__interfaces

    def getBases(self):
        """Get base classes.
        """
        klass = getattr(self.obj, '__class__', None)
        return getattr(klass, '__bases__', None)

    def getAttributes(self):
        """Get all attributes of an object.

        See objectinfo.txt to learn more.
        """
        klass = removeSecurityProxy(self.__klass)
        obj = removeSecurityProxy(self.obj)
        for name in dir(obj):
            value = getattr(obj, name, None)
            if value is None:
                continue
            if inspect.ismethod(value) or inspect.ismethoddescriptor(value):
                continue
            entry = {
                'name': name,
                'value': `value`,
                'value_linkable': self.isLinkable(value),
                'type' : type(value),
                # type_link is a very browser oriented data
                # element. Move it to a view?
                'type_link': self.getTypeLink(type(value)),
                'interface': utilities.getInterfaceForAttribute(
                                 name, klass=klass)
                }
            entry.update(utilities.getPermissionIds(
                name, getCheckerForInstancesOf(klass)))
            yield entry

    def getMethods(self):
        """Get all methods of an object.

        Get a list of dicts, describing the methods of an object. The
        dicts support keys ``name`` (name of method), ``signature``,
        ``doc`` (method's docstring or ``None``) and ``interface``
        (the interface, where the method was defined or ``None``).
        """
        klass = removeSecurityProxy(self.__klass)
        obj = removeSecurityProxy(self.obj)
        for name in dir(obj):
            value = getattr(obj, name, None)
            if value is None:
                continue
            if not (inspect.ismethod(value)
                    and not inspect.ismethoddescriptor(value)):
                continue
            if inspect.ismethod(value):
                signature = utilities.getFunctionSignature(value)
            else:
                signature = '(...)'

            entry = {
                'name': name,
                'signature' : signature,
                'doc' : getattr(value, '__doc__', None),
                'interface': utilities.getInterfaceForAttribute(
                                 name, klass=klass),
                'value_linkable': self.isLinkable(value),
                }                
            entry.update(utilities.getPermissionIds(
                name, getCheckerForInstancesOf(klass)))
            yield entry
        
    def isSequence(self):
        """Is the object observed a sequence?
        """
        if isinstance(self.obj, types.ListType):
            return True
        if isinstance(self.obj, types.TupleType):
            return True
        return IExtendedReadSequence.providedBy(self.obj)

    def getSequenceItems(self):
        """Get the items of a sequence.

        Returns a list of dicts, each representing one element of the
        sequence.
        """
        if not self.isSequence():
            return
        
        elems = []
        naked = removeSecurityProxy(self.obj)
        for index in xrange(0, len(self.obj)):
            value = naked[index]
            elems.append({
                'index' : index,
                'value' : value,
                'value_type' : type(value),
                'value_type_link' : self.getTypeLink(type(value)),
                'value_linkable': self.isLinkable(value),
                })
        return elems

    def isMapping(self):
        """Is the object observed a mapping?

        Mappings are those objects, which are dicts or provide
        IEnumerableMapping.
        """
        if isinstance(self.obj, types.DictType):
            return True
        return IEnumerableMapping.providedBy(self.obj)

    def getMappingItems(self):
        """Get the elements of a mapping.

        The elements are delivered as a list of dicts, each dict
        containing the keys ``key``, ``key_string`` (the key as a
        string), ``value``, ``value_type`` and ``value_type_link``.
        """
        elems = []
        naked = removeSecurityProxy(self.obj)
        if not hasattr(naked, 'items'):
            return []
        for key, value in naked.items():
            elems.append({
                'key' : key,
                'key_string' : `key`,
                'value' : value,
                'value_type' : type(value),
                'value_type_link' : self.getTypeLink(type(value)),
                'value_linkable': self.isLinkable(value),
                })
        return elems

    def isAnnotatable(self):
        """Does the object observed expect to be annotated?
        """
        return IAnnotatable.providedBy(self.obj)


    def getAnnotationsInfo(self):
        """Get all annotations associated with an object.

        If no annotations are associated with the object, ``None`` is
        returned. Otherwise we get a list of dicts, each containing
        keys ``key``, ``key_string`` (textual representation of key),
        ``value``, ``value_type`` and ``value_type_link''.
        """
        if not self.isAnnotatable():
            return []
        naked = removeSecurityProxy(self.obj)
        annotations = IAnnotations(naked)
        if not hasattr(annotations, 'items'):
            return []
        elems = []
        for key, value in annotations.items():
            elems.append({
                'key' : key,
                'key_string' : `key`,
                'value' : value,
                'value_type' : type(value),
                'value_type_link' : self.getTypeLink(type(value)),
                'value_linkable': self.isLinkable(value),
                'obj' : annotations[key]
                })
        return elems

    def getId(self):
        """Try to determine some kind of name.
        """
        return (getattr(self.obj, '__name__', None)
               or getattr(self.obj, 'id', None)
               or getattr(self.obj, '_o_id', None))




"""
We can use grok.AddForm to render an add form for objects:

  >>> getRootFolder()["zoo"] = Zoo()

  >>> from zope.testbrowser.testing import Browser
  >>> browser = Browser()
  >>> browser.handleErrors = False

  >>> browser.open("http://localhost/zoo/@@addmammoth")
  >>> browser.getControl(name="form.name").value = "Manfred the Mammoth"
  >>> browser.getControl(name="form.size").value = "Really big"
  >>> browser.getControl("Add entry").click()
  >>> print browser.contents
  Hi, my name is Manfred the Mammoth, and I\'m "Really big"

Instead of calling an object constructor with the form data, we can
also use the ``applyData`` method to store the data on the object.

  >>> browser.open("http://localhost/zoo/@@addmammothapplydata")
  >>> browser.getControl(name="form.name").value = "Ellie the Mammoth"
  >>> browser.getControl(name="form.size").value = "Really small"
  >>> browser.getControl("Add entry").click()
  >>> print browser.contents
  Hi, my name is Ellie the Mammoth, and I\'m "Really small"

"""
import grok
from zope import schema
from zope.interface import Interface, implements

class Zoo(grok.Container):
    pass

class IMammoth(Interface):
    name = schema.TextLine(title=u"Name")
    size = schema.TextLine(title=u"Size", default=u"Quite normal")

class Mammoth(grok.Model):
    implements(IMammoth)

    def __init__(self, name='', size=''):
        self.name = name
        self.size = size

class Index(grok.View):
    grok.context(Mammoth)
    def render(self):
        return 'Hi, my name is %s, and I\'m "%s"' % (self.context.name,
                                                     self.context.size)

class AddMammoth(grok.AddForm):
    grok.context(Zoo)

    form_fields = grok.AutoFields(Mammoth)

    @grok.action('Add entry')
    def add(self, **data):
        # pass data into Mammoth constructor
        self.context['manfred'] = manfred = Mammoth(**data)
        self.redirect(self.url(manfred))

class AddMammothApplyData(AddMammoth):

    @grok.action('Add entry')
    def add(self, **data):
        # instantiate Mammoth and then use self.applyData()
        self.context['ellie'] = ellie = Mammoth()
        self.applyData(ellie, **data)
        self.redirect(self.url(ellie))



"""
You can't call grok.name multiple times for a view

  >>> import grok.tests.view.namemultiple_fixture
  Traceback (most recent call last):
    ...
  GrokImportError: grok.name can only be called once per class.

"""



"""
A permission has to be defined first (using grok.Permission for example)
before it can be used in grok.require() in an XMLRPC class. This is even the
case for a default permission that is never used.

  >>> grok.testing.grok(__name__)
  Traceback (most recent call last):
   ...
  ConfigurationExecutionError: martian.error.GrokError: Undefined permission 'foo' in <class 'grok.tests.security.missing_permission_xmlrpc3.MissingPermission'>. Use grok.Permission first.
  ...

"""

import grok
import zope.interface

class Foo(grok.Permission):
    grok.name('foo')

class MissingPermission(grok.XMLRPC):
    grok.context(zope.interface.Interface)
    grok.require('doesnt.exist')

    @grok.require('foo')
    def foo(self):
        pass



# -*- coding: latin-1 -*-
"""
We do not accept bogus inline template such as ones that contain
encoded strings:

  >>> import grok
  >>> grok.PageTemplate('''
  ... <html>
  ... <body><h1 tal:content="string:Mammoth Cave Painting"/>
  ... <p>ööö</p>
  ... </body>
  ... </html>''')
  Traceback (most recent call last):
    ...
  ValueError: Invalid page template. Page templates must be unicode or ASCII.
"""



##############################################################################
#
# Copyright (c) 2007 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
=================================
DocGrok: a class browser for Grok
=================================

DocGrok offers an extensible class browser for packages, modules,
classes and similar things. To use it in your own application see
``docgrok.txt`` in the ``grok.admin`` package.

Here only the functionality as a class browser for the admin-UI is
covered.

Overview page
-------------

When we go to the documentation section, we should get an overview:

  >>> from zope.testbrowser.testing import Browser
  >>> browser = Browser()
  >>> browser.addHeader('Authorization', 'Basic mgr:mgrpw')
  >>> browser.open('http://localhost/')
  >>> browser.getLink('Documentation').click()
  >>> print browser.contents
  <html xmlns="http://www.w3.org/1999/xhtml">
  ... Welcome to DocGrok...

The overview offers us direct links to the zope package,

  >>> link = browser.getLink('browse the zope package')
  >>> link
  <Link text='browse the zope package' url='http://localhost/docgrok/zope'>

the grok package,

  >>> link = browser.getLink('browse the grok package')
  >>> link
  <Link text='browse the grok package' url='http://localhost/docgrok/grok'>

and a link to the internal object browser, which is different from the
class browser, and shows the ZODB root:

  >>> link = browser.getLink('ZODB root folder')
  >>> link
  <Link text='ZODB root folder' url='http://localhost/@@inspect.html'>
  
There are several things, that can be displayed by the class
browser. We start with packages.


DocGrok for packages
--------------------

We placed a package in the ``apackage`` directory. Let's try to fetch
documentation for it. We form a URL string, that contains the
package's dotted name with dots replaced by slashes:

  >>> pkg_dotted_name = __name__.rsplit('.',1)[0]
  >>> url_path = 'http://localhost/docgrok/%s/apackage' % (
  ...    pkg_dotted_name.replace('.', '/'))
  >>> browser.open(url_path)

Is it the documentation of the ``apackage`` package?

  >>> print browser.contents
  <html xmlns="http://www.w3.org/1999/xhtml">
  ...
  ...<span><a ...>.apackage</a></span>...
  ...
  ...(Python Package)...
  ...

Okay. In the page top we should have links to the various parent
packages contained in the dotted name of the examined package:

  >>> browser.getLink('grok')
  <Link text='grok' url='http://localhost/docgrok/grok'>

  >>> browser.getLink('ftests')
  <Link text='.ftests' url='http://localhost/docgrok/grok/ftests'>

and so on.


DocGrok for modules
-------------------

DocGrok for classes
-------------------

Custom DocGroks
---------------

It is relatively easy to provide own DocGroks, that are displayed when
browsing. You have to provide three parts:

* a DocGrok class

* a DocGrokHandlerClass

* a view for the DocGrok class

The latter is optional, but required, if you want to see your docgrok
documentation in the docgrok browser.

Basically, the DocGrok should provide and extract all information
available for the kind of thing (a class, a function or whatever), you
want to document. The handler class determines, whether a special
dotted path denotes one of the things you want to document while the
docgrok view finally renders all this (hopefully) nicer than it is
done below.

The *DocGrok* classes should be derived from
grok.admin.docgrok.DocGrok, so that they automatically provide some
useful information like the Python path of an object and similar.

The *DocGrokHandlers* should be derived from
grok.admin.docgrok.DocGrokHandler to be registered automatically on
startup. It must provide a method ``getDoctor(dotted_path,
obj=None)``, which returns a DocGrok object iff the given dotted_path
denotes a thing of the appropriate type. In the example below we
check, whether a given dotted path denotes a Mammoth class and in
this case return a ``DocGrokForMammoths`` instance.

The *DocGrokView* finally renders the information of the DocGrok
instance, which is created, when a user requests information about a
certain MammothManager.

Example:
--------

We created the wonderfull Mammoth as below. Normally, when we watch
the docgrok documentation of that class, we would be served the usual
class documentation. But we want to give hints, that mammoths are
really _large_, so we have to provide a special docgrok.

All three of the above mentioned parts are defined below: a docgrok
only for the Mammoth class (and its subclasses), a docgrok handler,
which finds out, whether something is a Mammoth class and a view to
display some valuable information about mammoths.

Both, the handler and the view are registered automatically on
startup, because they are subclassing DocGrokHandler (the handler) and
grok.View (the view). There is nothing else, we have to do.

To check this, we have a look at the docgrok class browser. First we
have a look at the module, the class is contained in::

  >>> browser.open("http://localhost/docgrok/grok/ftests/admin/docgrok")

In this page, there should be a link available to our Mammoth class,
as defined below::

  >>> link = browser.getLink(url='/Mammoth')
  >>> link.text
  'Mammoth'

  >>> link.url
  'http://localhost/docgrok/grok/ftests/admin/docgrok/Mammoth'
  
If we click on this link, we normally would get a usual class
documentation page as generated by the docgrok for classes. But, while
we have registered a special docgrok for Mammoth-things, we get::

  >>> link.click()
  >>> print browser.contents
  An enormous beast.
  Mammoths are really tall.
  The size is: remarkable


"""
import grok
from grok.admin.docgrok import DocGrok, DocGrokHandler
from grok.admin.view import DocGrokView

class Mammoth(object):
    """A (large) thing, we want to document later on.
    """
    pass

class DocGrokForMammoths(DocGrok):
    """Documentation for Mammoths.
    """
    def getSize(self):
        return u"remarkable"

class DocGrokForMammothsHandler(DocGrokHandler):
    """A class for determining, whether a dotted path denotes a
    Mammoth.
    """
    def getDoctor(self, dotted_path, ob=None):
        """The only method required from a docgrok handler.
        """
        from zope.dottedname.resolve import resolve
        try:
            ob = resolve(dotted_path)
        except ImportError:
            return
        try:
            if not issubclass(ob, Mammoth):
                return
        except TypeError:
            return
        return DocGrokForMammoths(dotted_path)

class DocGrokViewForMammoths(grok.View):
    """A view, that should fit into the other docgrok documentation.
    """
    # We bind to the docgrok which provides us with information about
    # the thing, we want to document.
    grok.context(DocGrokForMammoths)
    grok.name('index')
    
    def render(self):
        """To avoid an extra template, we provide a render method.
        """
        return (u"An enormous beast.\n"
                u"Mammoths are really tall.\n"
                u"The size is: %s " % self.context.getSize())




"""
Before a view is rendered, the update() method is executed. It can be
used e. g. to execute side effects or set up data for use in the
template.

  >>> grok.testing.grok(__name__)

We need to set up a default ITraversable adapter so that TALES
expressions can resolve paths:

  >>> from zope import component
  >>> from zope.traversing.adapters import DefaultTraversable
  >>> component.provideAdapter(DefaultTraversable, (None,))

  >>> manfred = Mammoth()
  >>> from zope.publisher.browser import TestRequest
  >>> request = TestRequest()
  >>> view = component.getMultiAdapter((manfred, request), name='cavepainting')
  >>> print view()
  <html>
  <body>
  <h1>red</h1>
  <h1>red</h1>
  </body>
  </html>
  

"""
import grok

class Mammoth(grok.Model):
    pass

class CavePainting(grok.View):
    def update(self):
        self.color = "red"


cavepainting = grok.PageTemplate("""\
<html>
<body>
<h1 tal:content="view/color"/>
<h1 tal:content="python: view.color"/>
</body>
</html>
""")



"""
When you call the grok.local_utility directive multiple times specifying
the same (interface, name) combination, we expect an error:

  >>> grok.testing.grok(__name__)
  Traceback (most recent call last):
    ...
  GrokError: Conflicting local utility registration
  <class 'grok.tests.utility.multiple_directive.Fireplace2'> in site
  <class 'grok.tests.utility.multiple_directive.Cave'>.
  Local utilities are registered multiple times for interface
  <InterfaceClass grok.tests.utility.multiple_directive.IFireplace> and
  name u''.  
"""
import grok
from zope import interface

class IFireplace(interface.Interface):
    pass

class Fireplace(grok.LocalUtility):
    grok.implements(IFireplace)

class Fireplace2(grok.LocalUtility):
    grok.implements(IFireplace)
    
class Cave(grok.Model, grok.Site):
    grok.local_utility(Fireplace, provides=IFireplace)
    grok.local_utility(Fireplace2, provides=IFireplace)



import os
import codecs
import sys
import urllib

import docutils.core
from docutils.writers.html4css1 import Writer

from zope.app.renderer.rest import ZopeTranslator
from zope.pagetemplate.pagetemplate import PageTemplate
from zope.pagetemplate.pagetemplatefile import PageTemplateFile

# registers code-block directive using pygments.
import pygments_code_block_directive

class ReStructuredTextToHTMLRenderer:
    """Convert from Restructured Text to HTML."""

    def __init__(self,content):
        self.content = content

    def render(self):
        settings_overrides = {
            'halt_level': 6,
            'input_encoding': 'utf8',
            'output_encoding': 'utf8',
            'initial_header_level': 2,
            # don't try to include the stylesheet (docutils gets hiccups)
            'stylesheet_path': '',
        }

        writer = Writer()
        writer.translator_class = ZopeTranslator
        html = docutils.core.publish_string(
                        self.content,
                        writer=writer,
                        settings_overrides=settings_overrides,)
        html = codecs.decode(html, 'utf_8')
        return html

class RestFile(object):

    source = ''
    target = ''
    url = ''
    title = ''
    active = True

    def __init__(self, url, source, target):
        self.url = url
        self.target = target
        if os.path.isfile(target) and os.path.isfile(source):
            if os.path.getmtime(source) < os.path.getmtime(target):
                self.active = False
        if os.path.isfile(source):
            self.source = codecs.open(source,"r",'utf8').read()
        elif source.startswith("http"):
            print "Downloading %s" % source
            try:
                response = urllib.urlopen(source)
                self.source = response.read()
            except IOError, e:
                if hasattr(e, 'reason'):
                    print 'We failed to reach a server.'
                    print 'Reason: ', e.reason
                elif hasattr(e, 'code'):
                    print 'The server couldn\'t fulfill the request.'
                    print 'Error code:', e.code
                self.source = u''
        else:
            self.source = source
        srctxt = self.source.split('\n')
        title = ''
        count = 0
        while title == '':
            title = srctxt[count]
            count += 1
            if title.startswith("=====") or title.startswith(".."):
                title = srctxt[count]
        self.title = title

    def set_rest_content(self):
        renderer = ReStructuredTextToHTMLRenderer(self.source)
        self.content = renderer.render().strip()

    def create_html(self, page, settings):
        if not self.active:
            pass
        print 'Processing ', self.url
        self.set_rest_content()

        settings["context"] = Context(self.url, self.title, self.content)
        content = page.pt_render(namespace=settings)
        #fp = open(self.target, 'w')
        targetdir = os.path.dirname(self.target)
        if not os.path.isdir(targetdir):
            print 'Creating dir: ', targetdir
            os.mkdir(targetdir)
        fp = codecs.open(self.target,"w",'utf8')
        fp.write(content)
        fp.close()

class Context:
    """Set up the context for rendering the rest as html through zpt"""

    def __init__(self, url, title=u'', content=u''):
        self.url = url
        self.title = title
        self.content = content

    @property
    def menu(self):
        for item in Menu:
            if item.get('href') == self.url:
                item['klass'] = u'selected'
            else:
                item['klass'] = u''
            # just for tutorial files
            if len(self.url.split('/')) > 2:
                if item.get('href').split('/')[:-1] == self.url.split('/')[:-1]:
                    item['klass'] = u'selected'
            if not item.get('description', None):
                item['description'] = item['title']
        return Menu

def create_html(rest_files, template):
    settings = {}
    settings["here"] = { "layout": template }
    page = PageTemplate()
    page.write("""<metal:block use-macro="here/layout/macros/pagelayout" />""")

    for restfile in rest_files:
        restfile.create_html(page, settings)

# Menu should later be generated if site becomes more complex
Menu = [
        {'href':'/index.html','title':u'Home','klass':''},
        {'href':'/about.html','title':u'About','klass':''},
        {'href':'/tutorial.html','title':u'Tutorial','klass':''},
        {'href':'/minitutorials/index.html','title':u'How Tos','klass':''},
        ]

def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    if not len(argv) == 1:
        print "Usage: grok2html OUTDIR"
        sys.exit(1)

    source_dir = os.path.dirname(__file__)
    www_dir = os.path.abspath(argv[0])

    if not os.path.isdir(www_dir):
        print "OUTDIR '%s' does not exist." % (www_dir,)
        sys.exit(1)

    os.chdir(source_dir)

    rest_files = []
    # Toplevel pages:
    for name in [
        'about',
        'index',
        'tutorial',
        'upgrade'
        ]:
        rest_files.append(
            RestFile(
                name, os.path.join(source_dir, name+'.txt'),
            os.path.join(www_dir, name+'.html')))
    # The minitutorials:
    for name in [
        'index',
        'macros',
        'permissions',
        'rest',
        'searching',
        'template-languages',
        'transient-objects',
        'xmlprc',
        ]:
        rest_files.append(
            RestFile(
                name, os.path.join(source_dir, 'minitutorials', name+'.txt'),
            os.path.join(www_dir, 'minitutorials', name+'.html')))
    # External sources:
    rest_files.append(
        RestFile(
            'zc.buildout',
            'http://svn.zope.org/*checkout*/zc.buildout/trunk/doc/tutorial.txt',
            os.path.join(www_dir, 'minitutorials', 'buildout.html')))

    template = PageTemplateFile(os.path.join(source_dir, 'template.pt'))
    create_html(rest_files, template)

if __name__ == '__main__':
    main()



##############################################################################
#
# Copyright (c) 2007 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Macros for the grok admin UI

  >>> from zope.testbrowser.testing import Browser
  >>> browser = Browser()
  >>> browser.addHeader('Authorization', 'Basic mgr:mgrpw')
  >>> browser.open('http://localhost/applications')

Check, that the macros template renders correctly, even if not called
with context of a GAIA object. This is important, because the macro
view is bound to ``Interface`` and can therefore be called with nearly
every object as context.

We create a non-GAIA object, a mammoth called 'manfred'.

  >>> subform = browser.getForm(name='Mammoth')
  >>> subform.getControl('Name your new app:').value = 'manfred'
  >>> subform.getControl('Create').click()

and call the macroview with it:

  >>> browser.open('http://localhost/manfred/@@externalview')
  >>> print browser.contents
  <html xmlns="http://www.w3.org/1999/xhtml">
  ...
     This template (grokadminmacros.pt in grok.admin) must be called
     from a view with defined root_url.
  ...

So developers get informed, that they called the wrong macro view.

Let's clean up.

  >>> browser.open('http://localhost/applications')
  >>> ctrl = browser.getControl(name='items')
  >>> ctrl.getControl(value='manfred').selected = True
  >>> browser.getControl('Delete Selected').click()


"""
import grok

class Mammoth(grok.Application, grok.Container):
    pass

class ExternalView(grok.View):
    """A view that calls grokadminmacros 'illegally'.
    """
    grok.context(Mammoth)

externalview = grok.PageTemplate("""\
<html metal:use-macro="context/@@grokadminmacros/macros/gaia-page">
</html>
""")

